# mypy: ignore-errors

"""
Status bar rendering for Pulka.

This module provides functions for rendering the status bar that displays
metadata about the current view including filename, row position, column info,
filters, sort, and memory usage.
"""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any

from ..utils import lazy_imports

if TYPE_CHECKING:
    StyleAndTextTuples = list[tuple[str, str]]  # type: ignore[assignment]
else:  # pragma: no cover - runtime import helper
    StyleAndTextTuples = lazy_imports.prompt_toolkit_style_and_text_tuples()

from ..core.formatting import _format_large_number_compact, _format_number_with_thousands_separator
from ..testing import is_test_mode
from .style_resolver import get_active_style_resolver
from .styles import apply_style

# Constants for formatting and truncation
_LARGE_NUMBER_THRESHOLD = 999999  # Threshold for compact number formatting
_MEMORY_THRESHOLD = 1000000  # Threshold for memory usage interpretation


def sample_memory_usage(*, test_mode: bool) -> int | None:
    """Return current memory usage in MB with the same semantics as the status bar."""

    if test_mode:
        return 120

    try:
        import resource
    except Exception:  # pragma: no cover - resource module unavailable
        return None

    try:
        usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    except Exception:  # pragma: no cover - getrusage failures are best-effort
        return None

    if usage > _MEMORY_THRESHOLD:
        return max(1, int(usage / 1024 / 1024))
    return max(1, int(usage / 1024))


def _simplify_dtype_text(dtype: Any) -> str:
    """Return a short, human-friendly representation of a Polars dtype."""

    text = str(dtype)
    for delimiter in ("(", "[", "<"):
        if delimiter in text:
            text = text.split(delimiter, 1)[0]
    # Normalize casing for common types
    return text.strip()


def _truncate_middle(text: str, max_length: int) -> str:
    """Return ``text`` truncated around the middle to fit ``max_length``."""

    if max_length <= 0:
        return ""
    if len(text) <= max_length:
        return text

    ellipsis = "…"
    if max_length <= len(ellipsis):
        return ellipsis[:max_length]

    slice_length = max_length - len(ellipsis)
    front = (slice_length + 1) // 2
    back = slice_length // 2
    return f"{text[:front]}{ellipsis}{text[-back:] if back else ''}"


def render_status_line(
    v: Any,
    *,
    test_mode: bool | None = None,
    resource_sample: int | None = None,
) -> StyleAndTextTuples:
    """Render a ``prompt_toolkit`` fragment representing the status bar."""

    if test_mode is None:
        test_mode = is_test_mode()

    if not hasattr(v, "columns") or not hasattr(v, "cur_col"):
        raise TypeError(
            f"Expected Viewer-like object with columns and cur_col attributes, got {type(v)}"
        )

    sheet = getattr(v, "sheet", None)
    columns = list(getattr(v, "columns", []) or [])
    if columns:
        cur_index = min(max(int(getattr(v, "cur_col", 0)), 0), len(columns) - 1)
        current_col = str(columns[cur_index])
        sheet_schema = getattr(sheet, "schema", {}) if sheet is not None else {}
        schema = getattr(v, "schema", None) or sheet_schema
        col_dtype = schema.get(columns[cur_index], "unknown")
    else:
        current_col = "no columns"
        col_dtype = "N/A"

    total_rows = getattr(v, "_total_rows", None)
    row_count_stale = bool(getattr(v, "_row_count_stale", False))
    was_pending = row_count_stale or total_rows is None
    if row_count_stale or total_rows is None:
        ensure_total_rows = getattr(v, "_ensure_total_rows", None)
        if callable(ensure_total_rows):
            total_rows = ensure_total_rows()
        elif hasattr(v, "sheet") and hasattr(v.sheet, "__len__"):
            with contextlib.suppress(Exception):
                total_rows = len(v.sheet)

    current_row_formatted = _format_number_with_thousands_separator(
        int(getattr(v, "cur_row", 0)) + 1
    )
    pending_row_count = getattr(v, "_row_count_future", None)
    display_pending = bool(getattr(v, "_row_count_display_pending", False))
    if total_rows is not None:
        sheet_id = getattr(getattr(v, "sheet", None), "sheet_id", None)
        if not (display_pending or (was_pending and sheet_id is not None)):
            if total_rows > _LARGE_NUMBER_THRESHOLD:
                rows_total_text = _format_large_number_compact(total_rows)
            else:
                rows_total_text = _format_number_with_thousands_separator(total_rows)
        else:
            rows_total_text = "≈"
    else:
        rows_total_text = "≈" if (pending_row_count is not None or was_pending) else "?"

    if resource_sample is None:
        mem_mb = sample_memory_usage(test_mode=test_mode)
    else:
        mem_mb = resource_sample

    is_file_browser = bool(getattr(sheet, "is_file_browser", False))
    browser_path: str | None = None
    if is_file_browser:
        browser_path = getattr(sheet, "display_path", None)
        if not browser_path:
            directory = getattr(sheet, "directory", None)
            if directory is not None:
                browser_path = str(directory)
        if not browser_path:
            fallback = getattr(sheet, "sheet_id", None) or getattr(v, "_source_path", None)
            if fallback is not None:
                browser_path = str(fallback)

    simple_dtype = _simplify_dtype_text(col_dtype)
    if browser_path:
        left_parts = [browser_path]
    else:
        left_parts = [f"row {current_row_formatted} / col {current_col}[{simple_dtype}]"]

    filter_text = getattr(v, "filter_text", None)
    if filter_text:
        left_parts.append(f"F: {filter_text}")

    sort_col = getattr(v, "sort_col", None)
    if sort_col:
        sort_asc = bool(getattr(v, "sort_asc", True))
        left_parts.append(f"S: {sort_col} ({'↑' if sort_asc else '↓'})")

    status_message = getattr(v, "status_message", None)
    if status_message:
        normalised_status = status_message.strip().lower()
        duplicate = False
        if filter_text:
            preview = filter_text if len(filter_text) <= 60 else f"{filter_text[:57]}..."
            filter_candidates = {
                filter_text.strip().lower(),
                preview.strip().lower(),
                f"filter: {filter_text.strip().lower()}",
                f"filter: {preview.strip().lower()}",
            }
            duplicate = normalised_status in filter_candidates
        if not duplicate and sort_col:
            arrow = "↑" if bool(getattr(v, "sort_asc", True)) else "↓"
            sort_candidates = {
                f"sort {sort_col} {arrow}".strip().lower(),
                f"sort {sort_col} ({arrow})".strip().lower(),
            }
            duplicate = normalised_status in sort_candidates
        if not duplicate:
            left_parts.append(status_message)

    left_text = " • ".join(left_parts)

    total_columns = len(columns)
    hidden_column_count = 0
    hidden_columns = getattr(v, "hidden_columns", None)
    if hidden_columns:
        hidden_column_count = len(tuple(hidden_columns))
    else:
        hidden_names = getattr(v, "_hidden_cols", None)
        if hidden_names and columns:
            hidden_set = set(hidden_names)
            hidden_column_count = sum(1 for name in columns if name in hidden_set)

    non_hidden_columns = total_columns - hidden_column_count
    if non_hidden_columns <= 0:
        visible_columns = list(getattr(v, "visible_cols", []) or [])
        visible_count = len(visible_columns)
        non_hidden_columns = visible_count or total_columns or visible_count

    right_parts = [f"depth {getattr(v, 'stack_depth', 0)}"]
    right_parts.append(f"{rows_total_text}×{max(0, non_hidden_columns)}")
    if mem_mb is not None:
        right_parts.append(f"mem {mem_mb}MB")
    right_text = " • ".join(right_parts)

    width = max(20, int(getattr(v, "view_width_chars", 80) or 0))
    right_text = _truncate_middle(right_text, width)
    available_left = max(0, width - len(right_text))
    if right_text and available_left > 0:
        available_left -= 1
    left_text = _truncate_middle(left_text, available_left) if available_left > 0 else ""

    if right_text:
        gap = max(1, width - len(left_text) - len(right_text))
        status = f"{left_text}{' ' * gap}{right_text}" if left_text else right_text.rjust(width)
    else:
        status = left_text

    status = status[:width]
    padded = status.ljust(width)
    if test_mode:
        style = ""
    else:
        resolver = get_active_style_resolver()
        style = resolver.prompt_toolkit_style_for_classes(("status",))
    return [(style, padded)]


def render_status_line_text(v: Any, *, test_mode: bool | None = None) -> str:
    """Return an ANSI string fallback for tests and non-PTK paths."""

    fragments = render_status_line(v, test_mode=test_mode)
    text = "".join(part for _, part in fragments)
    if not text:
        return text

    test_mode_flag = is_test_mode() if test_mode is None else test_mode
    resolver = get_active_style_resolver()
    components = resolver.resolve(("status",))
    style_str = components.to_prompt_toolkit() or None
    return apply_style(text, style_str, test_mode=test_mode_flag)



"""Engine-facing protocols and tabular data abstractions."""

from __future__ import annotations

import inspect
from collections.abc import Callable, Iterator, Mapping, Sequence, Sized
from dataclasses import dataclass, field
from types import MappingProxyType
from typing import Any, Protocol, cast, overload

from wcwidth import wcswidth, wcwidth  # type: ignore[import-untyped]

from ...utils import _get_int_env


@dataclass(slots=True, frozen=True)
class EnginePayloadHandle[PayloadT]:
    """Opaque wrapper describing an engine-specific payload."""

    engine: str
    kind: str
    _payload: PayloadT = field(repr=False)

    def unwrap(
        self,
        *,
        expected_engine: str | None = None,
        expected_kind: str | None = None,
    ) -> PayloadT:
        """Return the underlying payload after validating ``engine``/``kind``."""

        if expected_engine is not None and self.engine != expected_engine:
            msg = f"Handle is backed by engine '{self.engine}', not '{expected_engine}'"
            raise ValueError(msg)
        if expected_kind is not None and self.kind != expected_kind:
            msg = f"Handle is of kind '{self.kind}', not '{expected_kind}'"
            raise ValueError(msg)
        return self._payload

    def as_serializable(self) -> Mapping[str, str]:
        """Return metadata that can be safely serialised."""

        return {"engine": self.engine, "kind": self.kind}


_USE_ARRAY_BACKING = _get_int_env("PULKA_FF_ARRAY_SLICE", None, 1) != 0


def _array_len(array: Any) -> int:
    if isinstance(array, Sized):
        return len(array)
    if hasattr(array, "len"):
        return int(array.len())
    msg = f"Unsupported column backing without __len__: {type(array)!r}"
    raise TypeError(msg)


def _array_slice(array: Any, start: int, length: int | None = None) -> Any:
    length = None if length is None else max(0, length)
    slice_fn = getattr(array, "slice", None)
    if callable(slice_fn):
        slice_length = length if length is not None else _array_len(array) - start
        return slice_fn(start, slice_length)
    if hasattr(array, "__getitem__"):
        end = None if length is None else start + length
        return array[start:end]
    end = _array_len(array) if length is None else start + length
    return tuple(_array_get(array, idx) for idx in range(start, end))


def _array_get(array: Any, index: int) -> Any:
    item = getattr(array, "item", None)
    if callable(item):
        return item(index)
    if hasattr(array, "__getitem__"):
        return array[index]
    msg = f"Unsupported column backing without item access: {type(array)!r}"
    raise TypeError(msg)


def _array_iter(array: Any) -> Iterator[Any]:
    for idx in range(_array_len(array)):
        yield _array_get(array, idx)


def _array_null_count(array: Any) -> int:
    null_count_fn = getattr(array, "null_count", None)
    if callable(null_count_fn):
        try:
            return int(null_count_fn())
        except TypeError:
            pass
    is_null_fn = getattr(array, "is_null", None)
    if callable(is_null_fn):
        try:
            is_null = is_null_fn()
            sum_fn = getattr(is_null, "sum", None)
            if callable(sum_fn):
                return int(sum_fn())
        except Exception:  # pragma: no cover - best effort
            pass
    return sum(1 for value in _array_iter(array) if value is None)


def _concat_arrays(left: Any, right: Any) -> Any:
    try:  # pragma: no cover - Polars fast path
        import polars as pl

        if isinstance(left, pl.Series) and isinstance(right, pl.Series):
            return pl.concat([left, right], rechunk=False)
    except Exception:
        pass
    return tuple(_array_iter(left)) + tuple(_array_iter(right))


def _materialise_if_requested(array: Any) -> Any:
    if _USE_ARRAY_BACKING:
        return array
    return tuple(_array_iter(array))


def _maybe_legacy_formatter(
    candidate: Callable[..., Any] | None,
) -> Callable[[int], tuple[str, ...]] | None:
    if candidate is None:
        return None
    try:
        signature = inspect.signature(candidate)
    except (TypeError, ValueError):  # pragma: no cover - dynamic callables
        return None
    positional = [
        param
        for param in signature.parameters.values()
        if param.kind
        in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
    ]
    if len(positional) == 1:
        return cast(Callable[[int], tuple[str, ...]], candidate)
    return None


class DisplayCache:
    """Cache for lazily formatted display strings."""

    __slots__ = ("_data",)

    def __init__(self) -> None:
        self._data: dict[tuple[int, int, int | None], str] = {}

    def get(self, column_key: int, row: int, width: int | None) -> str | None:
        return self._data.get((column_key, row, width))

    def put(self, column_key: int, row: int, width: int | None, value: str) -> None:
        self._data[(column_key, row, width)] = value

    def clear_column(self, column_key: int) -> None:
        to_delete = [key for key in self._data if key[0] == column_key]
        for key in to_delete:
            self._data.pop(key, None)


def _display_width(value: str) -> int:
    width = wcswidth(value)
    if width < 0:
        return len(value)
    return int(width)


def _grapheme_clip(value: str, width: int) -> str:
    if width <= 0:
        return ""
    if _display_width(value) <= width:
        return value
    if width == 1:
        return "."
    current = 0
    pieces: list[str] = []
    for char in value:
        char_width = wcwidth(char)
        if char_width is None or char_width < 0:
            char_width = 1
        if current + char_width > width - 1:
            break
        pieces.append(char)
        current += char_width
    while pieces and current > width - 1:
        removed = pieces.pop()
        removed_width = wcwidth(removed)
        if removed_width is None or removed_width < 0:
            removed_width = 1
        current -= removed_width
    return "".join(pieces) + "."


def _format_scalar(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        try:
            return value.decode("utf-8", errors="replace")
        except Exception:  # pragma: no cover - defensive
            return value.decode("utf-8", errors="ignore")
    return str(value)


class ColumnValuesView(Sequence[Any]):
    __slots__ = ("_slice", "_column_index")

    def __init__(self, table_slice: TableSlice, column_index: int) -> None:
        self._slice = table_slice
        self._column_index = column_index

    def __len__(self) -> int:
        return self._slice.height

    def __getitem__(self, index: int | slice) -> Any:
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            return tuple(
                self._slice._get_value(self._column_index, idx) for idx in range(start, stop, step)
            )
        if index < 0:
            index += len(self)
        if index < 0 or index >= len(self):
            raise IndexError(index)
        return self._slice._get_value(self._column_index, index)

    def __iter__(self) -> Iterator[Any]:
        for idx in range(len(self)):
            yield self._slice._get_value(self._column_index, idx)

    def __eq__(self, other: object) -> bool:  # pragma: no cover - simple helper
        if isinstance(other, Sequence):
            return list(self) == list(other)
        return NotImplemented


class ColumnFormattedView(Sequence[str]):
    __slots__ = ("_slice", "_column_index", "_width")

    def __init__(self, table_slice: TableSlice, column_index: int, width: int) -> None:
        self._slice = table_slice
        self._column_index = column_index
        self._width = width

    def __len__(self) -> int:
        return self._slice.height

    @overload
    def __getitem__(self, index: int) -> str:  # pragma: no cover - signature only
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[str]:  # pragma: no cover - signature only
        ...

    def __getitem__(self, index: int | slice) -> str | Sequence[str]:
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            return tuple(
                self._slice._get_display(
                    self._column_index, idx, None if self._width == 0 else self._width
                )
                for idx in range(start, stop, step)
            )
        if index < 0:
            index += len(self)
        if index < 0 or index >= len(self):
            raise IndexError(index)
        return self._slice._get_display(
            self._column_index, index, None if self._width == 0 else self._width
        )

    def __iter__(self) -> Iterator[str]:
        for idx in range(len(self)):
            yield self._slice._get_display(
                self._column_index, idx, None if self._width == 0 else self._width
            )

    def __eq__(self, other: object) -> bool:  # pragma: no cover - simple helper
        if isinstance(other, Sequence):
            return list(self) == list(other)
        return NotImplemented


@dataclass(slots=True)
class TableColumn:
    """Column-oriented view over a tabular slice."""

    name: str
    data: Any
    dtype: Any
    null_count: int
    _display_fn: Callable[[int, int, Any, int | None], str] | None = field(default=None, repr=False)
    _owner: TableSlice | None = field(default=None, init=False, repr=False)
    _index: int = field(default=-1, init=False, repr=False)
    _values_view: ColumnValuesView | None = field(default=None, init=False, repr=False)
    _local_display_cache: dict[int, tuple[str, ...]] = field(
        default_factory=dict, init=False, repr=False
    )
    _local_values_cache: tuple[Any, ...] | None = field(default=None, init=False, repr=False)
    _legacy_formatter: Callable[[int], tuple[str, ...]] | None = field(
        default=None, init=False, repr=False
    )
    _legacy_display_cache: dict[int, tuple[str, ...]] = field(
        default_factory=dict, init=False, repr=False
    )
    _slice_offset: int = field(default=0, init=False, repr=False)

    def __post_init__(self) -> None:
        if not _USE_ARRAY_BACKING and not isinstance(self.data, tuple):
            object.__setattr__(self, "data", tuple(_array_iter(self.data)))
        legacy = _maybe_legacy_formatter(self._display_fn)
        if legacy is not None:
            self._legacy_formatter = legacy
            self._display_fn = None

    def _bind(self, owner: TableSlice, index: int) -> None:
        object.__setattr__(self, "_owner", owner)
        object.__setattr__(self, "_index", index)
        object.__setattr__(self, "_values_view", ColumnValuesView(owner, index))

    @property
    def values(self) -> Sequence[Any]:
        owner = self._owner
        if owner is None:
            cached = self._local_values_cache
            if cached is None:
                cached = tuple(_array_get(self.data, idx) for idx in range(len(self)))
                self._local_values_cache = cached
            return cached
        assert self._values_view is not None
        return self._values_view

    def formatted(self, max_chars: int) -> Sequence[str]:
        owner = self._owner
        if owner is None:
            width = max(max_chars, 0)
            if self._legacy_formatter is not None:
                cached = self._legacy_display_cache.get(width)
                if cached is None:
                    cached = self._legacy_formatter(width)
                    self._legacy_display_cache[width] = cached
                start = self._slice_offset
                end = start + len(self)
                return cached[start:end]
            cached = self._local_display_cache.get(width)
            if cached is not None:
                return cached
            base_values = self.values
            rendered = tuple(_format_scalar(value) for value in base_values)
            if width > 0:
                rendered = tuple(_grapheme_clip(text, width) for text in rendered)
            self._local_display_cache[width] = rendered
            return rendered
        safe_width = max(max_chars, 0)
        return ColumnFormattedView(owner, self._index, safe_width)

    def slice(self, start: int, length: int | None = None) -> TableColumn:
        data_slice = _array_slice(self.data, start, length)
        length_val = _array_len(data_slice) if length is None else max(0, length)
        null_count = _array_null_count(data_slice) if length_val else 0
        column = TableColumn(
            self.name,
            _materialise_if_requested(data_slice),
            self.dtype,
            null_count,
            self._display_fn,
        )
        column._legacy_formatter = self._legacy_formatter
        column._legacy_display_cache = self._legacy_display_cache
        column._slice_offset = self._slice_offset + start
        column._local_display_cache = self._local_display_cache
        column._local_values_cache = None
        return column

    def __len__(self) -> int:
        return _array_len(self.data)


@dataclass(slots=True)
class TableSlice:
    """Immutable window of table data returned to presentation layers."""

    columns: tuple[TableColumn, ...]
    schema: Mapping[str, Any]
    start_offset: int | None = None
    row_ids: Any | None = None
    _row_count: int = field(init=False, repr=False)
    _display_cache: DisplayCache = field(default_factory=DisplayCache, init=False, repr=False)
    _value_cache: dict[tuple[int, int], Any] = field(default_factory=dict, init=False, repr=False)

    def __post_init__(self) -> None:
        columns = tuple(self.columns)
        object.__setattr__(self, "columns", columns)
        schema_proxy = MappingProxyType(dict(self.schema))
        object.__setattr__(self, "schema", schema_proxy)
        for idx, column in enumerate(columns):
            column._bind(self, idx)

        if not columns:
            object.__setattr__(self, "_row_count", 0)
            return

        lengths = {len(column) for column in columns}
        if len(lengths) != 1:
            msg = "All columns must share the same row count"
            raise ValueError(msg)
        (row_count,) = lengths
        object.__setattr__(self, "_row_count", row_count)

    @classmethod
    def empty(
        cls,
        columns: Sequence[str] | None = None,
        schema: Mapping[str, Any] | None = None,
    ) -> TableSlice:
        base_schema = dict(schema or {})
        column_names = tuple(columns or base_schema.keys())
        table_columns: list[TableColumn] = []

        for name in column_names:
            dtype = base_schema.get(name)
            table_columns.append(TableColumn(name, (), dtype, 0))
            if name not in base_schema:
                base_schema[name] = dtype

        return cls(tuple(table_columns), base_schema)

    @property
    def column_names(self) -> tuple[str, ...]:
        return tuple(column.name for column in self.columns)

    @property
    def height(self) -> int:
        return self._row_count

    def __len__(self) -> int:
        return self.height

    def column(self, name: str) -> TableColumn:
        for column in self.columns:
            if column.name == name:
                return column
        msg = f"Unknown column: {name}"
        raise KeyError(msg)

    def column_at(self, index: int) -> TableColumn:
        return self.columns[index]

    def value_at(self, row: int, column_index: int) -> Any:
        return self._get_value(column_index, row)

    def get_cell(self, row: int, column_index: int, *, width: int | None = None) -> str:
        safe_width = None if width is None else max(width, 0)
        return self._get_display(column_index, row, safe_width)

    def iter_rows(self, start: int = 0, stop: int | None = None) -> Iterator[list[str]]:
        end = self.height if stop is None else min(stop, self.height)
        for row in range(start, end):
            yield [self.get_cell(row, col) for col in range(len(self.columns))]

    def slice(self, start: int, length: int | None = None) -> TableSlice:
        if length is None:
            sliced_columns = tuple(column.slice(start, None) for column in self.columns)
        else:
            sliced_columns = tuple(column.slice(start, length) for column in self.columns)
        row_ids = None
        if self.row_ids is not None:
            row_ids = _array_slice(self.row_ids, start, length)
        offset = None if self.start_offset is None else self.start_offset + start
        return TableSlice(sliced_columns, self.schema, start_offset=offset, row_ids=row_ids)

    def view(self, row_start: int, row_len: int) -> TableSlice:
        return self.slice(row_start, row_len)

    def concat_vertical(self, other: TableSlice) -> TableSlice:
        if self.column_names != other.column_names:
            msg = "Table slices must share the same column ordering"
            raise ValueError(msg)
        merged: list[TableColumn] = []
        for left, right in zip(self.columns, other.columns, strict=True):
            combined = _concat_arrays(left.data, right.data)
            null_count = left.null_count + right.null_count
            merged.append(
                TableColumn(
                    left.name,
                    _materialise_if_requested(combined),
                    left.dtype,
                    null_count,
                    left._display_fn,
                )
            )
            merged[-1]._legacy_formatter = left._legacy_formatter
            merged[-1]._legacy_display_cache = left._legacy_display_cache
            merged[-1]._slice_offset = left._slice_offset
        row_ids = None
        if self.row_ids is not None or other.row_ids is not None:
            left_ids = self.row_ids if self.row_ids is not None else ()
            right_ids = other.row_ids if other.row_ids is not None else ()
            row_ids = _concat_arrays(left_ids, right_ids)
        start_offset = self.start_offset
        if start_offset is None and other.start_offset is not None:
            candidate = other.start_offset - len(self)
            if candidate >= 0:
                start_offset = candidate
        return TableSlice(
            tuple(merged),
            self.schema,
            start_offset=start_offset,
            row_ids=row_ids,
        )

    def _absolute_row(self, row: int) -> int:
        if self.start_offset is None:
            return row
        return self.start_offset + row

    def _get_value(self, column_index: int, row: int) -> Any:
        abs_row = self._absolute_row(row)
        column = self.columns[column_index]
        key = (id(column.data), abs_row)
        cached = self._value_cache.get(key)
        if cached is not None:
            return cached
        value = _array_get(column.data, row)
        self._value_cache[key] = value
        return value

    def _get_display(self, column_index: int, row: int, width: int | None) -> str:
        column = self.columns[column_index]
        abs_row = self._absolute_row(row)
        key_width = width
        cached = self._display_cache.get(id(column.data), abs_row, key_width)
        if cached is not None:
            return cached
        raw_value = self._get_value(column_index, row)
        if column._legacy_formatter is not None:
            width_key = max(width or 0, 0)
            formatted = column._legacy_display_cache.get(width_key)
            if formatted is None:
                formatted = column._legacy_formatter(width_key)
                column._legacy_display_cache[width_key] = formatted
            idx = column._slice_offset + row
            text = formatted[idx] if 0 <= idx < len(formatted) else ""
            if width is not None:
                text = _grapheme_clip(text, width)
            self._display_cache.put(id(column.data), abs_row, key_width, text)
            return text
        display_fn = column._display_fn
        if display_fn is not None:
            display_row = row + column._slice_offset
            text = display_fn(display_row, abs_row, raw_value, width)
            if width is not None:
                text = _grapheme_clip(text, width)
        else:
            text = _format_scalar(raw_value)
            if width is not None:
                text = _grapheme_clip(text, width)
        self._display_cache.put(id(column.data), abs_row, key_width, text)
        return text


PhysicalPlanHandle = EnginePayloadHandle[Any]
"""Engine-neutral handle representing a compiled physical plan."""

# Backwards compatibility for legacy imports that still refer to ``PhysicalPlan``.
PhysicalPlan = PhysicalPlanHandle


class EngineAdapter(Protocol):
    """Compile logical plans and validate predicates for a storage engine."""

    def compile(self, plan: Any) -> PhysicalPlan: ...

    def validate_filter(self, clause: str) -> None: ...


class Materializer(Protocol):
    """Collect physical plans into :class:`TableSlice` objects."""

    def collect(self, plan: PhysicalPlan) -> TableSlice: ...

    def collect_slice(
        self,
        plan: PhysicalPlan,
        *,
        start: int = 0,
        length: int | None = None,
        columns: Sequence[str] | None = None,
    ) -> TableSlice: ...

    def collect_slice_stream(
        self,
        plan: PhysicalPlan,
        *,
        start: int = 0,
        length: int | None = None,
        columns: Sequence[str] | None = None,
        batch_rows: int | None = None,
    ) -> Iterator[TableSlice]: ...

    def count(self, plan: PhysicalPlan) -> int | None: ...


__all__ = [
    "EnginePayloadHandle",
    "EngineAdapter",
    "Materializer",
    "PhysicalPlan",
    "PhysicalPlanHandle",
    "TableColumn",
    "TableSlice",
]



"""prompt_toolkit-native table control that renders from a viewport plan."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from contextlib import nullcontext, suppress
from dataclasses import dataclass, replace
from time import perf_counter
from types import MethodType
from typing import TYPE_CHECKING, Any
from weakref import ReferenceType, ref

from prompt_toolkit.data_structures import Point
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.layout.controls import UIContent, UIControl

from ...logging import Recorder, frame_hash, viewer_state_snapshot
from ...render.status_bar import render_status_line
from ...render.style_resolver import get_active_style_resolver
from ...render.table import (
    RenderedLine,
    build_blank_line,
    build_row_line,
    build_separator_line,
    compute_column_overflows,
    determine_blank_line_highlights,
)
from ...render.viewport_plan import Cell, ViewportPlan, compute_viewport_plan
from ...theme import theme_epoch

PlanCacheKey = tuple[Any, int, int, int, int, Any, int, int, int]
LineCacheKey = tuple[Any, ...]

if TYPE_CHECKING:  # pragma: no cover - import cycles in typing context only
    from ...core.viewer import Viewer
else:  # pragma: no cover - fallback for runtime typing introspection
    Viewer = Any


@dataclass(slots=True)
class _LineRender:
    """Container for rendered line fragments and plain text."""

    fragments: StyleAndTextTuples
    plain_text: str
    cursor_x: int | None = None
    source: RenderedLine | None = None


@dataclass(frozen=True, slots=True)
class _BudgetPlan:
    """Per-frame rendering adjustments when the budget guard triggers."""

    overscan_hint: int | None = None
    minimal_styles: bool = False
    drop_borders: bool = False
    coalesce_multiplier: float = 1.0

    def is_default(self) -> bool:
        return (
            self.overscan_hint is None
            and not self.minimal_styles
            and not self.drop_borders
            and self.coalesce_multiplier == 1.0
        )


class RenderBudgetGuard:
    """Track frame timings and request temporary render degradations."""

    def __init__(self, *, budget_ms: float = 18.0, alpha: float = 0.25) -> None:
        self._budget_ms = budget_ms
        self._alpha = alpha
        self._avg_ms: float | None = None
        self._degrade_next = False
        self._sample_count = 0

    def plan_frame(self, viewer: Viewer) -> _BudgetPlan:
        if not self._degrade_next:
            return _BudgetPlan()

        self._degrade_next = False
        overscan = max(1, int(getattr(viewer, "view_height", 1) or 1))
        return _BudgetPlan(
            overscan_hint=overscan,
            minimal_styles=True,
            drop_borders=True,
            coalesce_multiplier=2.0,
        )

    def record_frame(self, render_ms: float, paint_ms: float) -> None:
        total = render_ms + paint_ms
        self._sample_count += 1
        if self._avg_ms is None:
            self._avg_ms = total
        else:
            self._avg_ms = (1 - self._alpha) * self._avg_ms + self._alpha * total

        if self._sample_count >= 2 and (
            total > self._budget_ms or (self._avg_ms or 0.0) > self._budget_ms
        ):
            self._degrade_next = True

    @property
    def average_ms(self) -> float | None:
        return self._avg_ms


class TableControl(UIControl):
    """A prompt_toolkit ``UIControl`` that draws table cells from a viewport plan."""

    def __init__(
        self,
        viewer: Viewer,
        *,
        apply_pending_moves: Callable[[], None],
        poll_background_jobs: Callable[[], None],
        set_status: Callable[[StyleAndTextTuples], None],
        apply_budget_plan: Callable[[_BudgetPlan], None] | None = None,
        recorder: Recorder | None = None,
    ) -> None:
        self._viewer = viewer
        self._apply_pending_moves = apply_pending_moves
        self._poll_background_jobs = poll_background_jobs
        self._set_status = set_status
        self._apply_budget_plan = apply_budget_plan
        self._recorder = recorder
        self._cached_plan: ViewportPlan | None = None
        self._plan_cache_key: PlanCacheKey | None = None
        self._line_cache: dict[LineCacheKey, _LineRender] = {}
        self._wrapped_ui_hooks: set[Any] = set()
        self._wrapped_ui_hook_refs: list[ReferenceType[Any]] = []
        self._wrapped_ui_hook_ids: set[int] = set()
        self._budget_guard = RenderBudgetGuard()
        self._current_budget_plan = _BudgetPlan()
        self._current_theme_epoch = theme_epoch()
        self._recorded_style_epoch: int | None = None
        self._last_plan_cache_hit = False
        self._last_frame_finished_at: float | None = None
        self._last_move_direction: int = 0
        self._last_move_time: float | None = None
        self._last_coalesced_steps: int = 0
        self._last_overscan_rows: int = 0
        self._coalesce_hot_window = 0.25
        self._overscan_ratio = 0.25

    def update_viewer(self, viewer: Viewer) -> None:
        """Point the control at the active viewer."""

        self._viewer = viewer
        self._invalidate_cache()

    def attach_ui_hooks(self, ui_hooks: Any) -> Any:
        """Wrap ``ui_hooks.invalidate`` to clear cached render state."""

        invalidate = getattr(ui_hooks, "invalidate", None)
        if invalidate is None:
            return ui_hooks

        hook_id = id(ui_hooks)

        try:
            if ui_hooks in self._wrapped_ui_hooks:
                return ui_hooks
        except TypeError:
            pass

        for hook_ref in list(self._wrapped_ui_hook_refs):
            hook = hook_ref()
            if hook is None:
                self._wrapped_ui_hook_refs.remove(hook_ref)
                continue
            if hook is ui_hooks:
                return ui_hooks

        if hook_id in self._wrapped_ui_hook_ids:
            return ui_hooks

        bound_self = getattr(invalidate, "__self__", None)
        bound_func = getattr(invalidate, "__func__", None)

        def _wrapped_invalidate(*args: Any, **kwargs: Any) -> Any:
            self._invalidate_cache()
            if bound_self is not None and bound_func is not None:
                return bound_func(bound_self, *args[1:], **kwargs)
            return invalidate(*args[1:], **kwargs) if args else invalidate(**kwargs)

        ui_hooks.invalidate = MethodType(_wrapped_invalidate, ui_hooks)
        try:
            self._wrapped_ui_hooks.add(ui_hooks)
        except TypeError:
            try:
                self._wrapped_ui_hook_refs.append(ref(ui_hooks))
            except TypeError:
                self._wrapped_ui_hook_ids.add(hook_id)
        return ui_hooks

    def preferred_width(self, max_available_width: int) -> int | None:  # noqa: D401
        return None

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: Callable[[int], str] | None,
    ) -> int | None:  # noqa: D401
        return None

    def is_focusable(self) -> bool:  # noqa: D401
        return True

    def create_content(self, width: int, height: int) -> UIContent:  # noqa: D401
        viewer = self._viewer
        recorder = self._recorder if self._recorder and self._recorder.enabled else None

        move_controller = getattr(self._apply_pending_moves, "__self__", None)
        base_steps = (
            max(1, int(getattr(move_controller, "_max_steps_per_frame", 1)))
            if move_controller is not None
            else 1
        )
        pending_before = (
            abs(int(getattr(move_controller, "_pending_row_delta", 0)))
            if move_controller is not None
            else 0
        )
        previous_direction = self._last_move_direction
        burst_multiplier = self._compute_burst_multiplier(
            pending_before, base_steps, previous_direction
        )

        budget_plan = self._budget_guard.plan_frame(viewer)
        if burst_multiplier > getattr(budget_plan, "coalesce_multiplier", 1.0):
            budget_plan = replace(budget_plan, coalesce_multiplier=burst_multiplier)
        self._current_budget_plan = budget_plan
        previous_theme_epoch = self._current_theme_epoch
        self._current_theme_epoch = theme_epoch()
        if previous_theme_epoch != self._current_theme_epoch:
            self._recorded_style_epoch = None
        if self._apply_budget_plan is not None:
            self._apply_budget_plan(budget_plan)
        if budget_plan.overscan_hint is not None:
            with suppress(Exception):
                viewer.request_frame_budget_overscan(budget_plan.overscan_hint)

        direction, _total_steps, coalesced_steps = self._apply_pending_moves_with_coalesce(
            budget_plan,
            previous_direction,
            base_steps,
            move_controller,
        )
        self._poll_background_jobs()
        overscan_rows = self._maybe_prime_overscan(
            viewer,
            height,
            budget_plan,
            direction,
            previous_direction,
            move_controller,
        )

        perf_ctx = (
            recorder.perf_timer(
                "render.table",
                payload={"context": "tui", "trigger": "refresh"},
            )
            if recorder
            else nullcontext()
        )

        render_start = perf_counter()
        with perf_ctx:
            plan, plan_cache_hit = self._get_plan(viewer, width, height)
            rendered, line_stats = self._render_lines(plan, height)
        render_end = perf_counter()

        status_ctx = (
            recorder.perf_timer(
                "render.status",
                payload={"context": "tui", "trigger": "refresh"},
            )
            if recorder
            else nullcontext()
        )
        status_fragments: StyleAndTextTuples = []
        with status_ctx:
            status_fragments = render_status_line(viewer)
        self._set_status(status_fragments)
        viewer.acknowledge_status_rendered()
        status_text = "".join(fragment for _, fragment in status_fragments)

        if recorder:
            state_snapshot = viewer_state_snapshot(viewer)
            recorder.record_state(state_snapshot)
            if status_text:
                recorder.record_status(status_text)

        if recorder:
            self._maybe_record_line_styles(rendered, recorder)

        fragments: list[StyleAndTextTuples] = [line.fragments for line in rendered]
        plain_lines = [line.plain_text for line in rendered]

        cursor_position = self._locate_cursor(rendered)

        content = UIContent(
            get_line=lambda line_index: fragments[line_index],
            line_count=len(fragments),
            cursor_position=cursor_position,
            show_cursor=True,
        )

        paint_end = perf_counter()

        if recorder and not budget_plan.is_default():
            recorder.record(
                "render_budget",
                {
                    "avg_ms": self._budget_guard.average_ms,
                    "render_ms": (render_end - render_start) * 1000,
                    "paint_ms": (paint_end - render_end) * 1000,
                },
            )

        self._budget_guard.record_frame(
            (render_end - render_start) * 1000,
            (paint_end - render_end) * 1000,
        )

        if recorder:
            frame_lines = plain_lines[:]
            if status_text:
                frame_lines.append(status_text)
            frame_capture = "\n".join(frame_lines)
            recorder.record_frame(
                frame_text=frame_capture,
                frame_hash=frame_hash(frame_capture),
            )
            total_bytes = sum(len(line) for line in plain_lines)
            recorder.record(
                "render_stats",
                {
                    "component": "table_control",
                    "lines": len(rendered),
                    "rendered_lines": line_stats["rendered"],
                    "reused_lines": line_stats["reused"],
                    "plan_cache": "hit" if plan_cache_hit else "miss",
                    "plain_bytes": total_bytes,
                    "coalesced_steps": coalesced_steps,
                    "overscan_rows": overscan_rows,
                },
            )

        self._last_plan_cache_hit = plan_cache_hit
        self._last_frame_finished_at = paint_end
        self._last_coalesced_steps = coalesced_steps
        self._last_overscan_rows = overscan_rows

        try:
            return content
        finally:
            self._current_budget_plan = _BudgetPlan()
            if self._apply_budget_plan is not None:
                self._apply_budget_plan(_BudgetPlan())

    def _render_lines(
        self, plan: ViewportPlan, height: int
    ) -> tuple[list[_LineRender], dict[str, int]]:
        column_widths = [max(1, column.width) for column in plan.columns] or [1]
        frozen_boundary = plan.frozen_boundary_idx
        column_overflows = compute_column_overflows(plan.columns, plan.rows > 0)
        column_widths_key = tuple(column_widths)
        column_overflows_key = tuple(column_overflows)
        cache_hits = 0
        cache_misses = 0
        used_keys: set[LineCacheKey] = set()
        highlight_top_blank, highlight_bottom_blank = determine_blank_line_highlights(plan)

        def _cache_lookup(key: LineCacheKey, builder: Callable[[], _LineRender]) -> _LineRender:
            nonlocal cache_hits, cache_misses
            cached = self._line_cache.get(key)
            if cached is not None:
                cache_hits += 1
                used_keys.add(key)
                return cached
            cache_misses += 1
            line = builder()
            self._line_cache[key] = line
            used_keys.add(key)
            return line

        def _blank_line(*, header: bool = False, row_active: bool = False) -> _LineRender:
            key = self._blank_line_key(
                column_widths_key,
                frozen_boundary,
                column_overflows_key,
                header=header,
                row_active=row_active,
            )
            return _cache_lookup(
                key,
                lambda: self._to_line_render(
                    build_blank_line(
                        column_widths,
                        frozen_boundary,
                        column_overflows,
                        header=header,
                        column_plans=plan.columns,
                        row_active=row_active,
                    )
                ),
            )

        lines: list[_LineRender] = []

        table_height = max(0, height - 1)
        has_header = bool(plan.cells and plan.cells[0] and plan.cells[0][0].role == "header")
        body_rows = plan.cells[1:] if has_header else plan.cells
        drop_borders = self._current_budget_plan.drop_borders
        include_boundary = not drop_borders or plan.frozen_boundary_idx is not None

        if table_height > 0:
            lines.append(_blank_line(header=has_header, row_active=highlight_top_blank))

        if has_header:
            header_cells = plan.cells[0]
            header_key = self._row_line_key(
                header_cells,
                column_widths_key,
                column_overflows_key,
                frozen_boundary,
                is_header=True,
                include_boundary=include_boundary,
            )
            lines.append(
                _cache_lookup(
                    header_key,
                    lambda cells=header_cells: self._to_line_render(
                        build_row_line(
                            cells,
                            column_widths,
                            frozen_boundary,
                            column_overflows,
                            is_header=True,
                            row_active=False,
                            column_plans=plan.columns,
                            include_boundary=include_boundary,
                        )
                    ),
                )
            )
            if body_rows:
                sep_key = self._separator_line_key(column_widths_key)
                lines.append(
                    _cache_lookup(
                        sep_key,
                        lambda: self._to_line_render(build_separator_line(column_widths)),
                    )
                )

        for row in body_rows:
            row_key = self._row_line_key(
                row,
                column_widths_key,
                column_overflows_key,
                frozen_boundary,
                is_header=False,
                include_boundary=include_boundary,
            )
            lines.append(
                _cache_lookup(
                    row_key,
                    lambda row=row: self._to_line_render(
                        build_row_line(
                            row,
                            column_widths,
                            frozen_boundary,
                            column_overflows,
                            is_header=False,
                            column_plans=plan.columns,
                            include_boundary=include_boundary,
                        )
                    ),
                )
            )

        if height > 0:
            lines.append(_blank_line(row_active=highlight_bottom_blank))

        stats = {"reused": cache_hits, "rendered": cache_misses}
        current_cache = self._line_cache
        self._line_cache = {key: current_cache[key] for key in used_keys if key in current_cache}
        return lines, stats

    def _locate_cursor(self, lines: list[_LineRender]) -> Point:
        line_index = 0
        cursor_x = 0
        for idx, line in enumerate(lines):
            cx = getattr(line, "cursor_x", None)
            if cx is not None:
                line_index = idx
                cursor_x = cx
                break
        return Point(x=cursor_x, y=line_index)

    def _to_line_render(self, line: RenderedLine) -> _LineRender:
        fragments: StyleAndTextTuples = []
        for segment in line.segments:
            style = self._style_from_classes(segment.classes)
            fragments.append((style, segment.text))
        return _LineRender(fragments, line.plain_text, line.cursor_x, line)

    def _style_from_classes(self, classes: Sequence[str]) -> str:
        filtered = self._filter_classes(classes)
        if not filtered:
            return ""
        resolver = get_active_style_resolver()
        style = resolver.prompt_toolkit_style_for_classes(filtered)
        return style or ""

    def _filter_classes(self, classes: Sequence[str]) -> tuple[str, ...]:
        if not self._current_budget_plan.minimal_styles:
            return tuple(classes)

        essential = {
            "table",
            "table.header",
            "table.cell",
            "table.cell.null",
            "table.row.active",
            "table.cell.active",
            "table.col.active",
            "table.header.active",
            "table.header.sorted",
            "table.separator",
        }
        return tuple(cls for cls in classes if cls in essential)

    def _maybe_record_line_styles(self, lines: Sequence[_LineRender], recorder: Recorder) -> None:
        if self._recorded_style_epoch == self._current_theme_epoch:
            return

        resolver = get_active_style_resolver()
        captured: list[dict[str, Any]] = []

        for index, line in enumerate(lines):
            source = line.source
            if source is None:
                continue
            if not any("table.header" in segment.classes for segment in source.segments):
                continue

            segments_payload: list[dict[str, Any]] = []
            for segment in source.segments:
                components = resolver.resolve(segment.classes)
                segments_payload.append(
                    {
                        "text": segment.text,
                        "classes": list(segment.classes),
                        "foreground": components.foreground,
                        "background": components.background,
                        "extras": list(components.extras),
                    }
                )

            if segments_payload:
                captured.append(
                    {
                        "line_index": index,
                        "plain_text": source.plain_text,
                        "segments": segments_payload,
                    }
                )
            break

        if not captured:
            return

        recorder.record_render_line_styles(
            component="table_control",
            lines=captured,
            metadata={"theme_epoch": self._current_theme_epoch},
        )
        self._recorded_style_epoch = self._current_theme_epoch

    def _get_plan(self, viewer: Viewer, width: int, height: int) -> tuple[ViewportPlan, bool]:
        key_hint = self._make_plan_key(viewer, width, height)
        if self._cached_plan is not None and self._plan_cache_key == key_hint:
            return self._cached_plan, True

        plan = compute_viewport_plan(viewer, width, height)
        epoch = self._current_epoch(viewer)
        view_id = self._view_identity(viewer)
        self._cached_plan = plan
        self._plan_cache_key = (
            view_id,
            plan.row_offset,
            plan.col_offset,
            width,
            height,
            epoch,
            getattr(viewer, "cur_row", 0),
            getattr(viewer, "cur_col", 0),
            self._current_theme_epoch,
        )
        # Column level cache must be invalidated when plan changes.
        self._line_cache = {}
        return plan, False

    def _make_plan_key(self, viewer: Viewer, width: int, height: int) -> PlanCacheKey:
        view_id = self._view_identity(viewer)
        row_offset = self._estimate_row_offset(viewer)
        col_offset = max(0, getattr(viewer, "col0", 0))
        epoch = self._current_epoch(viewer)
        cur_row = getattr(viewer, "cur_row", 0)
        cur_col = getattr(viewer, "cur_col", 0)
        theme_ep = getattr(self, "_current_theme_epoch", theme_epoch())
        return (
            view_id,
            row_offset,
            col_offset,
            width,
            height,
            epoch,
            cur_row,
            cur_col,
            theme_ep,
        )

    def _estimate_row_offset(self, viewer: Viewer) -> int:
        row0 = max(0, getattr(viewer, "row0", 0))
        frozen_rows = getattr(viewer, "visible_frozen_row_count", 0)
        if isinstance(frozen_rows, int) and frozen_rows > 0:
            row0 = max(row0, frozen_rows)
        return row0

    def _current_epoch(self, viewer: Viewer) -> Any:
        sheet = getattr(viewer, "sheet", None)
        if sheet is not None:
            version = getattr(sheet, "cache_version", None)
            if version is not None:
                return version
        generation_getter = getattr(viewer, "job_generation", None)
        if callable(generation_getter):
            try:
                return generation_getter()
            except Exception:  # pragma: no cover - defensive
                return None
        return None

    def _view_identity(self, viewer: Viewer) -> Any:
        sheet_id = getattr(viewer, "sheet_id", None)
        if sheet_id is not None:
            return sheet_id
        return id(viewer)

    def _invalidate_cache(self) -> None:
        self._cached_plan = None
        self._plan_cache_key = None
        self._line_cache = {}

    def _row_line_key(
        self,
        cells: list[Cell],
        column_widths: tuple[int, ...],
        column_overflows: tuple[bool, ...],
        frozen_boundary: int | None,
        *,
        is_header: bool,
        include_boundary: bool,
    ) -> LineCacheKey:
        cell_key: tuple[Any, ...] = tuple(
            (
                cell.text,
                cell.active_row,
                cell.active_col,
                cell.active_cell,
                cell.role,
            )
            for cell in cells
        )
        return (
            "row",
            is_header,
            include_boundary,
            column_widths,
            frozen_boundary,
            column_overflows,
            cell_key,
            self._current_theme_epoch,
            self._current_budget_plan.minimal_styles,
            self._current_budget_plan.drop_borders,
        )

    def _blank_line_key(
        self,
        column_widths: tuple[int, ...],
        frozen_boundary: int | None,
        column_overflows: tuple[bool, ...],
        *,
        header: bool,
        row_active: bool,
    ) -> LineCacheKey:
        return (
            "blank",
            header,
            row_active,
            column_widths,
            frozen_boundary,
            column_overflows,
            self._current_theme_epoch,
            self._current_budget_plan.minimal_styles,
            self._current_budget_plan.drop_borders,
        )

    def _separator_line_key(self, column_widths: tuple[int, ...]) -> LineCacheKey:
        return (
            "separator",
            column_widths,
            self._current_theme_epoch,
            self._current_budget_plan.drop_borders,
        )

    def _apply_pending_moves_with_coalesce(
        self,
        budget_plan: _BudgetPlan,
        previous_direction: int,
        base_steps: int,
        controller: Any | None,
    ) -> tuple[int, int, int]:
        viewer = self._viewer
        before_row = getattr(viewer, "cur_row", 0)

        self._apply_pending_moves()

        after_row = getattr(viewer, "cur_row", before_row)
        delta = after_row - before_row
        direction = 0
        if delta > 0:
            direction = 1
        elif delta < 0:
            direction = -1
        consumed = abs(delta)
        total_moved = consumed
        coalesced = 0

        if (
            direction != 0
            and controller is not None
            and self._is_previous_frame_hot(previous_direction)
        ):
            pending_remaining = getattr(controller, "_pending_row_delta", 0)
            pending_same_dir = (
                max(0, int(pending_remaining)) if direction > 0 else max(0, -int(pending_remaining))
            )
            if pending_same_dir > 0:
                multiplier = float(getattr(budget_plan, "coalesce_multiplier", 1.0) or 1.0)
                multiplier = min(4.0, max(1.0, multiplier))
                max_allowed = max(base_steps, int(round(base_steps * multiplier)))
                remaining_budget = max(0, max_allowed - consumed)
                if remaining_budget > 0:
                    move_fn = getattr(viewer, "move_down" if direction > 0 else "move_up", None)
                    if callable(move_fn):
                        before_extra_row = after_row
                        move_fn(min(pending_same_dir, remaining_budget))
                        new_row = getattr(viewer, "cur_row", before_extra_row)
                        actual_extra = new_row - after_row if direction > 0 else after_row - new_row
                        actual_extra = max(0, actual_extra)
                        if actual_extra > 0:
                            total_moved += actual_extra
                            coalesced = actual_extra
                            if direction > 0:
                                with suppress(Exception):
                                    controller._pending_row_delta = max(
                                        0,
                                        int(getattr(controller, "_pending_row_delta", 0))
                                        - actual_extra,
                                    )
                            else:
                                with suppress(Exception):
                                    controller._pending_row_delta = min(
                                        0,
                                        int(getattr(controller, "_pending_row_delta", 0))
                                        + actual_extra,
                                    )

        if total_moved > 0:
            self._last_move_direction = direction
            self._last_move_time = perf_counter()
        else:
            self._last_move_direction = 0
        self._last_coalesced_steps = coalesced
        return direction, total_moved, coalesced

    def _compute_burst_multiplier(
        self, pending_rows: int, base_steps: int, previous_direction: int
    ) -> float:
        if base_steps <= 0:
            return 1.0
        if pending_rows <= base_steps:
            return 1.0
        if not self._is_previous_frame_hot(previous_direction):
            return 1.0
        blocks = (pending_rows + base_steps - 1) // base_steps
        extra = max(0, min(3, blocks - 1))
        if extra <= 0:
            return 1.0
        return 1.0 + float(extra)

    def _is_previous_frame_hot(self, previous_direction: int) -> bool:
        if previous_direction == 0:
            return False
        last_finish = self._last_frame_finished_at
        if last_finish is None:
            return False
        return (perf_counter() - last_finish) <= self._coalesce_hot_window

    def _maybe_prime_overscan(
        self,
        viewer: Viewer,
        height: int,
        budget_plan: _BudgetPlan,
        direction: int,
        previous_direction: int,
        controller: Any | None,
    ) -> int:
        max_extra_from_budget: int | None = None
        if budget_plan.overscan_hint is not None:
            try:
                hint_value = int(budget_plan.overscan_hint)
            except (TypeError, ValueError):  # pragma: no cover - defensive
                hint_value = 0
            body_height = max(0, height - 1)
            max_extra_from_budget = max(0, hint_value - body_height)
        hot_motion = self._is_previous_frame_hot(previous_direction)
        if not hot_motion:
            # Allow immediate same-direction scrolls even if the previous frame fell
            # outside the hot window (e.g. under CI load) so overscan still primes.
            same_direction = direction != 0 and direction == previous_direction
            if not same_direction:
                return 0
        if direction == 0:
            pending = (
                abs(int(getattr(controller, "_pending_row_delta", 0)))
                if controller is not None
                else 0
            )
            if pending <= 0:
                return 0
        body_height = max(0, height - 1)
        if body_height <= 0:
            return 0
        overscan_rows = int(body_height * self._overscan_ratio)
        if overscan_rows <= 0 and body_height > 0:
            overscan_rows = 1
        overscan_rows = min(overscan_rows, body_height * 3)
        if max_extra_from_budget is not None:
            overscan_rows = min(overscan_rows, max_extra_from_budget)
        if overscan_rows <= 0:
            return 0
        columns = self._visible_column_names(viewer)
        if not columns:
            return 0
        try:
            viewer.get_visible_table_slice(columns, overscan_hint=body_height + overscan_rows)
        except Exception:  # pragma: no cover - defensive
            return 0
        return overscan_rows

    def _visible_column_names(self, viewer: Viewer) -> list[str]:
        columns = getattr(viewer, "visible_cols", None)
        if columns:
            return list(columns)
        columns = getattr(viewer, "columns", None)
        if columns:
            return list(columns)
        return []



"""Background job helpers for column insight calculations."""

from __future__ import annotations

import math
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from time import perf_counter_ns
from typing import Any

import polars as pl

from ..column_insight import (
    ColumnInsight,
    InsightHistogram,
    TopValue,
    summarize_value_preview,
)
from ..formatting import _is_string_dtype, _supports_min_max, _supports_numeric_stats

_MAX_TOP_VALUES = 10
_HISTOGRAM_BIN_COUNT = 48


@dataclass(frozen=True, slots=True)
class ColumnInsightJobConfig:
    """Parameters captured when scheduling a column insight job."""

    column_name: str
    plan_hash: str | None
    sheet_id: str | None
    source_path: str | None = None
    top_values: int = _MAX_TOP_VALUES


def compute_column_insight(
    *,
    lazyframe: pl.LazyFrame,
    config: ColumnInsightJobConfig,
    schema: Mapping[str, pl.DataType] | None = None,
) -> ColumnInsight:
    """Compute metrics for ``config.column_name`` using ``lazyframe``."""

    start_ns = perf_counter_ns()
    column_name = config.column_name
    dtype = _resolve_dtype(column_name, schema, lazyframe)

    if dtype is None and column_name not in lazyframe.columns:
        duration = perf_counter_ns() - start_ns
        return ColumnInsight(
            config.sheet_id,
            config.plan_hash,
            column_name,
            None,
            None,
            None,
            None,
            None,
            None,
            {},
            (),
            None,
            source_path=config.source_path,
            duration_ns=duration,
            error=f"unknown column: {column_name}",
        )

    try:
        stats_frame = lazyframe.select(_build_aggregations(column_name, dtype)).collect()
    except Exception as exc:  # pragma: no cover - upstream polars failures
        duration = perf_counter_ns() - start_ns
        return ColumnInsight(
            config.sheet_id,
            config.plan_hash,
            column_name,
            _dtype_str(dtype),
            None,
            None,
            None,
            None,
            None,
            {},
            (),
            None,
            source_path=config.source_path,
            duration_ns=duration,
            error=str(exc),
        )

    stats_row = stats_frame.to_dicts()[0] if stats_frame.height else {}
    row_count = _to_int(stats_row.get("row_count"))
    non_null_count = _to_int(stats_row.get("non_null_count"))
    null_count = _to_int(stats_row.get("null_count"))
    distinct_count = _to_int(stats_row.get("distinct_count"))
    null_fraction = _fraction(null_count, row_count)

    stats_payload = _filter_stats(stats_row)
    top_value_limit = max(0, config.top_values) if _supports_top_value_summary(dtype) else 0
    top_values = _collect_top_values(
        lazyframe,
        column_name,
        limit=top_value_limit,
        non_null_count=non_null_count,
    )
    histogram = _collect_histogram(
        lazyframe,
        column_name,
        dtype=dtype,
        stats_row=stats_row,
        non_null_count=non_null_count,
    )

    duration = perf_counter_ns() - start_ns
    return ColumnInsight(
        config.sheet_id,
        config.plan_hash,
        column_name,
        _dtype_str(dtype),
        row_count,
        non_null_count,
        null_count,
        null_fraction,
        distinct_count,
        stats_payload,
        top_values,
        histogram,
        source_path=config.source_path,
        duration_ns=duration,
    )


def _build_aggregations(column: str, dtype: pl.DataType | None) -> Sequence[pl.Expr]:
    col_expr = pl.col(column)
    aggs: list[pl.Expr] = [
        pl.len().alias("row_count"),
        col_expr.is_not_null().sum().alias("non_null_count"),
        col_expr.is_null().sum().alias("null_count"),
        col_expr.n_unique().alias("distinct_count"),
    ]

    if dtype is not None and _supports_min_max(dtype):
        aggs.extend(
            [
                col_expr.min().alias("min"),
                col_expr.max().alias("max"),
            ]
        )

    if dtype is not None and _supports_numeric_stats(dtype):
        aggs.extend(
            [
                col_expr.mean().alias("mean"),
                col_expr.median().alias("median"),
                col_expr.quantile(0.95, interpolation="nearest").alias("p95"),
                col_expr.quantile(0.05, interpolation="nearest").alias("p05"),
                col_expr.std().alias("std"),
            ]
        )

    return aggs


def _collect_top_values(
    lazyframe: pl.LazyFrame,
    column: str,
    *,
    limit: int,
    non_null_count: int | None,
) -> tuple[TopValue, ...]:
    if limit <= 0 or not non_null_count:
        return ()

    try:
        top_frame = (
            lazyframe.select(pl.col(column))
            .drop_nulls()
            .group_by(column)
            .agg(pl.len().alias("count"))
            .sort("count", descending=True)
            .limit(limit)
            .collect()
        )
    except Exception:
        return ()

    values: list[TopValue] = []
    for row in top_frame.to_dicts():
        count = _to_int(row.get("count")) or 0
        raw_value = row.get(column)
        display, truncated = summarize_value_preview(raw_value, max_chars=48)
        fraction = _fraction(count, non_null_count)
        values.append(TopValue(raw_value, display, count, fraction, truncated))
    return tuple(values)


def _collect_histogram(
    lazyframe: pl.LazyFrame,
    column: str,
    *,
    dtype: pl.DataType | None,
    stats_row: Mapping[str, Any],
    non_null_count: int | None,
) -> InsightHistogram | None:
    if dtype is None or not _supports_numeric_stats(dtype):
        return None
    if non_null_count in (None, 0):
        return None
    minimum = stats_row.get("min")
    maximum = stats_row.get("max")
    if minimum is None or maximum is None:
        return None
    try:
        min_value = float(minimum)
        max_value = float(maximum)
    except Exception:
        return None
    if not (math.isfinite(min_value) and math.isfinite(max_value)):
        return None
    if max_value < min_value:
        min_value, max_value = max_value, min_value
    if math.isclose(max_value, min_value):
        bins = tuple(1.0 if idx == 0 else 0.0 for idx in range(_HISTOGRAM_BIN_COUNT))
        return InsightHistogram(bins)
    bin_count = _HISTOGRAM_BIN_COUNT
    width = (max_value - min_value) / bin_count
    if width <= 0 or not math.isfinite(width):
        return None
    col_expr = pl.col(column)
    try:
        bin_expr = ((col_expr - min_value) / width).floor()
        binned = (
            lazyframe.select(
                pl.when(col_expr.is_null())
                .then(None)
                .otherwise(
                    pl.when(bin_expr < 0)
                    .then(0)
                    .when(bin_expr >= bin_count)
                    .then(bin_count - 1)
                    .otherwise(bin_expr)
                    .cast(pl.Int64)
                )
                .alias("__bin")
            )
            .drop_nulls("__bin")
            .group_by("__bin")
            .agg(pl.len().alias("count"))
            .collect()
        )
    except Exception:
        return None
    counts = [0] * bin_count
    for row in binned.iter_rows():
        idx, count = row
        try:
            pos = int(idx)
        except Exception:
            continue
        if 0 <= pos < bin_count:
            try:
                counts[pos] = int(count)
            except Exception:
                continue
    max_count = max(counts) if counts else 0
    if max_count <= 0:
        return None
    normalized = tuple(count / max_count for count in counts)
    return InsightHistogram(normalized)


def _supports_top_value_summary(dtype: pl.DataType | None) -> bool:
    if dtype is None:
        return False
    if _is_string_dtype(dtype):
        return True
    is_categorical = getattr(pl.datatypes, "is_categorical", None)
    if is_categorical is not None:
        try:
            if is_categorical(dtype):
                return True
        except Exception:
            pass
    dtype_name = type(dtype).__name__.lower()
    return "categorical" in dtype_name or "enum" in dtype_name


def _filter_stats(row: Mapping[str, Any]) -> dict[str, Any]:
    skip = {"row_count", "non_null_count", "null_count", "distinct_count"}
    return {key: row.get(key) for key in row if key not in skip}


def _fraction(numerator: int | None, denominator: int | None) -> float | None:
    if numerator is None or denominator in (None, 0):
        return None
    try:
        assert numerator is not None
        assert denominator is not None
        return float(numerator) / float(denominator)
    except Exception:
        return None


def _to_int(value: Any) -> int | None:
    if value is None:
        return None
    try:
        return int(value)
    except Exception:
        return None


def _resolve_dtype(
    column: str,
    schema: Mapping[str, pl.DataType] | None,
    lazyframe: pl.LazyFrame,
) -> pl.DataType | None:
    if schema and column in schema:
        return schema[column]
    try:
        collected = lazyframe.schema
        if column in collected:
            return collected[column]
    except Exception:  # pragma: no cover - schema fetch may fail on older polars
        pass
    try:
        inferred = lazyframe.collect_schema()
        return inferred.get(column)
    except Exception:
        return None


def _dtype_str(dtype: pl.DataType | None) -> str | None:
    if dtype is None:
        return None
    try:
        text = str(dtype)
    except Exception:
        return dtype.__class__.__name__
    simple = dtype.__class__.__name__
    if text.startswith(f"{simple}("):
        return simple
    return text


__all__ = [
    "ColumnInsightJobConfig",
    "compute_column_insight",
]



# mypy: ignore-errors

"""Viewer state management for Pulka.

The heavy lifting is delegated to small helper components housed in
``plan_controller``, ``row_count_tracker`` and ``state``. ``Viewer`` now wires
those helpers together while exposing a stable orchestration surface for the
rest of the application.
"""

from __future__ import annotations

import contextlib
import math
import weakref
from collections import deque
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from time import monotonic_ns
from typing import TYPE_CHECKING, Any, Literal

import polars as pl

from ...config.settings import STREAMING_DEFAULTS
from ...sheets.data_sheet import DataSheet
from ...sheets.transformation_history import (
    SupportsSnapshots,
    TransformationHistory,
    TransformationSnapshot,
)
from ...testing import is_test_mode
from ..engine.contracts import TableSlice
from ..engine.viewer_engine import ViewerEngine
from ..errors import (
    CancelledError,
    CompileError,
    MaterializeError,
    PlanError,
    PulkaCoreError,
)
from ..interfaces import JobRunnerProtocol
from ..plan import QueryPlan
from ..plan_ops import reset as plan_reset
from ..plan_ops import set_filter as plan_set_filter
from ..plan_ops import set_projection as plan_set_projection
from ..plan_ops import set_sql_filter as plan_set_sql_filter
from ..plan_ops import toggle_sort as plan_toggle_sort
from ..row_provider import RowProvider
from ..sheet import SHEET_FEATURE_PLAN, sheet_supports
from .components import ColumnWidthController, FreezePaneController, RowCacheController
from .plan_controller import PlanController
from .row_count_tracker import RowCountTracker
from .state import ViewerSnapshot, ViewerStateController
from .transformation_manager import ChangeResult, ViewerTransformationManager
from .ui_hooks import NullViewerUIHooks, ViewerUIHooks


@dataclass(frozen=True, slots=True)
class ViewerCursor:
    """Immutable cursor position exposed to consumers."""

    row: int
    col: int


@dataclass(frozen=True, slots=True)
class ViewerViewport:
    """Immutable viewport bounds and span."""

    row0: int
    rowN: int  # noqa: N815 - keep mixedCase for parity with public snapshot
    col0: int
    colN: int  # noqa: N815 - keep mixedCase for parity with public snapshot
    rows: int
    cols: int


@dataclass(frozen=True, slots=True)
class ViewerPublicState:
    """Public, immutable snapshot of viewer state."""

    cursor: ViewerCursor
    viewport: ViewerViewport
    columns: tuple[str, ...]
    visible_columns: tuple[str, ...]
    hidden_columns: tuple[str, ...]
    frozen_columns: tuple[str, ...]
    total_rows: int | None
    visible_row_count: int
    total_columns: int
    visible_column_count: int
    hidden_column_count: int
    status_message: str | None
    highlighted_column: str | None
    width_mode: Literal["default", "single", "all"]
    width_target: int | None
    all_columns_maximized: bool
    sort_column: str | None
    sort_ascending: bool
    filter_text: str | None
    filter_kind: Literal["expr", "sql"] | None
    search_text: str | None
    frequency_mode: bool
    frequency_source_column: str | None

    @property
    def width_mode_state(self) -> dict[str, int | None | str]:
        """Return a serialisable representation of the active width mode."""

        return {"mode": self.width_mode, "target": self.width_target}


if TYPE_CHECKING:
    from ...api.session import Session
    from ..sheet import Sheet
else:
    Sheet = object
    Session = object


class Viewer:
    def __init__(
        self,
        sheet: Sheet,
        *,
        viewport_rows: int | None = None,
        viewport_cols: int | None = None,
        source_path: str | None = None,
        session: Session | None = None,
        row_provider: RowProvider | None = None,
        ui_hooks: ViewerUIHooks | None = None,
        runner: JobRunnerProtocol,
    ):
        self.sheet = sheet
        self.columns: list[str] = list(sheet.columns)
        self._schema_cache = getattr(sheet, "schema", {})
        self._source_path = source_path
        self._session_ref: weakref.ReferenceType[Session] | None = (
            weakref.ref(session) if session is not None else None
        )
        self._runner: JobRunnerProtocol = runner

        # Cursor & viewport
        self.cur_row = 0
        self.cur_col = 0
        self.row0 = 0
        self.col0 = 0
        self._total_rows: int | None = None
        self._row_count_stale: bool = True
        self._row_count_future = None
        self._row_count_display_pending: bool = False
        self._max_visible_col: int | None = None  # Limit rightmost visible column when set
        self._status_dirty: bool = False
        self._status_message: str | None = None

        # Terminal metrics
        self._viewport_rows_override = viewport_rows
        self._viewport_cols_override = viewport_cols
        # Column width heuristics - sample-based dynamic allocation
        self._min_col_width = 4
        self._default_col_width_cap = 20  # Cap for all columns except last visible
        self._sep_overhead = 3
        # Column hiding functionality
        self._hidden_cols: set[str] = set()  # effective hidden column names
        self._local_hidden_cols: set[str] = set()  # legacy cache when no plan is present

        self._view_width_override_chars: int | None = None

        # Controllers for modular responsibilities
        self._streaming_enabled_default = STREAMING_DEFAULTS.enabled
        self._streaming_batch_rows_default = STREAMING_DEFAULTS.batch_rows

        self._freeze = FreezePaneController(self)
        self._row_cache = RowCacheController(
            self,
            self._freeze,
            streaming_enabled=self._streaming_enabled_default,
            streaming_batch_rows=self._streaming_batch_rows_default,
        )
        self._widths = ColumnWidthController(self)

        # UI integration bridge (prompt_toolkit, headless, etc.).
        self._ui_hooks: ViewerUIHooks = ui_hooks or NullViewerUIHooks()

        provider = row_provider or getattr(sheet, "row_provider", None)
        if provider is None:
            provider = RowProvider.for_sheet(sheet, runner=self._runner)
        elif getattr(provider, "_runner", None) is None:
            with contextlib.suppress(Exception):
                provider._runner = self._runner  # type: ignore[attr-defined]
        self._row_provider = provider
        self._engine = ViewerEngine(self._row_provider)

        self._header_widths = self._compute_initial_column_widths()
        self._default_header_widths = list(self._header_widths)  # baseline to revert to
        self._width_mode: Literal["default", "single", "all"] = "default"
        self._width_target: int | None = None
        self._width_cache_all: list[int] | None = None
        self._width_cache_single: dict[int, int] = {}
        self._autosized_widths: dict[int, int] = {}
        self._sticky_column_widths: dict[str, int] = {}
        self._decimal_alignment_cache: dict[str, tuple[int, int]] = {}

        self._state = ViewerStateController(self)
        self._transformations = self._create_transformation_manager(sheet)
        self._row_counts = RowCountTracker(self, runner=self._runner)
        self._plan_controller = PlanController(self)

        self.update_terminal_metrics()
        # Cache for visible columns calculation
        self._visible_key: tuple[int, int, int | None] | None = None
        self._visible_cols_cached: list[str] = self.columns[:1] if self.columns else []
        self._aligning_active_column: bool = False

        # Row velocity tracking for adaptive overscan
        self._row0_velocity_samples: deque[tuple[int, int]] = deque(maxlen=6)
        self._last_row0_sample: int = self.row0
        self._last_row0_ns: int | None = None
        self._last_velocity_event_ns: int | None = None
        self._local_filter_text: str | None = None
        self._local_filter_kind: Literal["expr", "sql"] | None = None
        self._local_search_text: str | None = None
        self.is_freq_view: bool = False
        self.freq_source_col: str | None = None
        # Histogram: track whether this viewer hosts the numeric histogram sheet.
        self.is_hist_view: bool = False
        self._perf_callback: Callable[[str, float, dict[str, Any]], None] | None = None

        # Track the viewer's position within the sheet stack (0 = root dataset).
        self.stack_depth: int = 0
        self._frame_budget_overscan_hint: int | None = None

        self._apply_sheet_freeze_defaults(sheet)
        self._apply_file_browser_layout_defaults(sheet)
        self._sync_hidden_columns_from_plan()

    @property
    def sheet_id(self) -> str | None:
        """Expose the sheet identifier when available."""

        return getattr(self.sheet, "sheet_id", None)

    @property
    def job_runner(self) -> JobRunnerProtocol:
        return self._runner

    @property
    def row_provider(self) -> RowProvider:
        """Return the service responsible for fetching row slices."""

        return self._row_provider

    @property
    def engine(self) -> ViewerEngine:
        """Return the viewer engine responsible for data access."""

        return self._engine

    @property
    def state_controller(self) -> ViewerStateController:
        """Expose the controller that manages cursor and viewport state."""

        return self._state

    @property
    def plan_controller(self) -> PlanController:
        """Expose the controller that manages query plan mutations."""

        return self._plan_controller

    @property
    def row_count_tracker(self) -> RowCountTracker:
        """Expose the tracker responsible for refreshing row counts."""

        return self._row_counts

    def job_generation(self) -> int:
        """Return the sheet generation tracked by the job runner."""

        context = getattr(self.sheet, "job_context", None)
        if context is None:
            return 0
        _, generation, _ = context()
        return generation

    def plan_hash(self) -> str | None:
        """Return the current plan hash for job coalescing."""

        context = getattr(self.sheet, "job_context", None)
        if context is None:
            return None
        _, _, plan_hash = context()
        return plan_hash

    def _current_plan(self) -> Any:
        """Return the current plan object when available."""

        return self._plan_controller.current_plan()

    def _plan_projection_columns(self) -> tuple[str, ...] | None:
        """Return the active plan projection constrained to known columns."""

        plan = self._current_plan()
        if not isinstance(plan, QueryPlan):
            return None

        projection = plan.projection_or(self.columns)
        if not projection:
            return tuple(self.columns)

        known = set(self.columns)
        return tuple(name for name in projection if name in known)

    def _plan_compiler_for_validation(self) -> Any:
        """Return a plan compiler suitable for validating plan mutations."""

        return self._plan_controller.plan_compiler_for_validation()

    @property
    def session(self) -> Session | None:
        """Return the owning session when available."""

        if self._session_ref is None:
            return None
        return self._session_ref()

    @property
    def ui_hooks(self) -> ViewerUIHooks:
        """Return the UI hook bridge active for this viewer."""

        return self._ui_hooks

    def set_ui_hooks(self, hooks: ViewerUIHooks | None) -> None:
        """Swap the active UI hooks and refresh terminal metrics."""

        self._ui_hooks = hooks or NullViewerUIHooks()
        with contextlib.suppress(Exception):
            self.update_terminal_metrics()

    @property
    def schema(self) -> dict[str, pl.DataType]:
        """Expose the current schema, delegating to the underlying sheet."""
        schema = getattr(self.sheet, "schema", None)
        if schema is None:
            return {}
        self._schema_cache = schema
        return schema

    @property
    def filter_kind(self) -> Literal["expr", "sql"] | None:
        """Return the kind of active filter tracked by the current plan."""

        plan = self._current_plan()
        if plan is None:
            return self._local_filter_kind
        if plan.sql_filter is not None:
            return "sql"
        if plan.filters:
            return "expr"
        return None

    @property
    def filter_text(self) -> str | None:
        """Return the human readable filter description for the active plan."""

        plan = self._current_plan()
        if plan is None:
            return self._local_filter_text
        if plan.sql_filter is not None:
            return f"SQL WHERE {plan.sql_filter}"
        if plan.filters:
            return plan.filters[0]
        return None

    @filter_text.setter
    def filter_text(self, value: str | None) -> None:
        self._local_filter_text = value
        self._local_filter_kind = None
        if value is not None:
            prefix = "SQL WHERE "
            self._local_filter_kind = "sql" if value.startswith(prefix) else "expr"

    @property
    def sql_filter_text(self) -> str | None:
        """Return the raw SQL WHERE clause when an SQL filter is active."""

        plan = self._current_plan()
        if plan is None:
            if self._local_filter_kind == "sql" and self._local_filter_text:
                prefix = "SQL WHERE "
                if self._local_filter_text.startswith(prefix):
                    return self._local_filter_text[len(prefix) :]
                return self._local_filter_text
            return None
        return plan.sql_filter

    @property
    def search_text(self) -> str | None:
        """Return the active search text tracked by the plan when available."""

        if self._local_search_text is not None:
            return self._local_search_text
        plan = self._current_plan()
        if plan is None:
            return None
        return plan.search_text

    @search_text.setter
    def search_text(self, value: str | None) -> None:
        self._local_search_text = value

    @property
    def sort_col(self) -> str | None:
        """Expose the primary sort column derived from the current plan."""

        plan = self._current_plan()
        if plan is None or not plan.sort:
            return None
        return plan.sort[0][0]

    @property
    def sort_asc(self) -> bool:
        """Expose whether the primary sort column is ascending."""

        plan = self._current_plan()
        if plan is None or not plan.sort:
            return True
        return not plan.sort[0][1]

    def set_perf_callback(
        self,
        callback: Callable[[str, float, dict[str, Any]], None] | None,
    ) -> None:
        """Register a lightweight perf callback for internal hotspots."""
        self._perf_callback = callback

    def _record_perf_event(
        self,
        phase: str,
        duration_ms: float,
        payload: dict[str, Any],
    ) -> None:
        if not self._perf_callback:
            return
        with contextlib.suppress(Exception):
            self._perf_callback(phase, duration_ms, payload)

    def invalidate_row_cache(self) -> None:
        """Drop the cached row window used for fast vertical scrolling."""
        self._row_cache.invalidate()

    def _get_row_cache_prefetch(self) -> int:
        return self._row_cache.get_prefetch()

    # ------------------------------------------------------------------
    # Freeze panes helpers

    def _invalidate_frozen_columns_cache(self) -> None:
        self._freeze.invalidate_cache()

    def _apply_sheet_freeze_defaults(self, sheet: Sheet | None = None) -> None:
        target = self.sheet if sheet is None else sheet
        if target is None:
            return

        default_cols = getattr(target, "default_frozen_columns", None)
        if isinstance(default_cols, int) and default_cols >= 0:
            self.set_frozen_columns(default_cols)

        default_rows = getattr(target, "default_frozen_rows", None)
        if isinstance(default_rows, int) and default_rows >= 0:
            self.set_frozen_rows(default_rows)

    def _apply_file_browser_layout_defaults(self, sheet: Sheet | None = None) -> None:
        target = self.sheet if sheet is None else sheet
        if target is None or not getattr(target, "is_file_browser", False):
            return
        if not self.columns:
            return

        try:
            name_idx = self.columns.index("name")
        except ValueError:
            name_idx = 0

        self._width_mode = "single"
        self._width_target = name_idx
        self._apply_width_mode()

    def _ensure_frozen_columns_cache(self) -> None:
        self._freeze.ensure_cache()

    def _frozen_column_indices(self) -> list[int]:
        return self._freeze.column_indices()

    def _first_scrollable_col_index(self) -> int:
        return self._freeze.first_scrollable_col_index()

    def _is_column_frozen(self, idx: int) -> bool:
        return self._freeze.is_column_frozen(idx)

    @property
    def frozen_column_count(self) -> int:
        return self._freeze.column_count

    @property
    def frozen_row_count(self) -> int:
        return self._freeze.row_count

    def _effective_frozen_row_count(self) -> int:
        return self._freeze.effective_row_count()

    def _reserved_frozen_rows(self) -> int:
        """Return how many rows at the top of the viewport are occupied by frozen rows."""

        return self._freeze.reserved_view_rows()

    def _body_view_height(self) -> int:
        """Return how many rows are available for the scrollable body."""

        return self._freeze.body_view_height()

    def _max_row0_for_total(self, total_rows: int) -> int:
        """Return the largest valid ``row0`` for a dataset with ``total_rows`` rows."""

        return self._state.max_row0_for_total(total_rows)

    @property
    def frozen_columns(self) -> list[str]:
        return self._freeze.frozen_column_names()

    def _frozen_column_index_set(self) -> frozenset[int]:
        return self._freeze.column_index_set()

    def _frozen_column_name_set(self) -> frozenset[str]:
        return self._freeze.column_name_set()

    @property
    def visible_row_positions(self) -> list[int]:
        return self._row_cache.visible_row_positions()

    @property
    def visible_frozen_row_count(self) -> int:
        return self._row_cache.visible_frozen_row_count()

    def set_frozen_columns(self, count: int) -> None:
        self._freeze.set_frozen_columns(count)

    def set_frozen_rows(self, count: int) -> None:
        self._freeze.set_frozen_rows(count)

    def clear_freeze(self) -> None:
        self._freeze.clear()

    @property
    def hidden_columns(self) -> list[str]:
        """Return the columns currently hidden from the table."""

        if not self._hidden_cols:
            return []
        hidden: list[str] = []
        hidden_set = self._hidden_cols
        for name in self.columns:
            if name in hidden_set:
                hidden.append(name)
        return hidden

    def _update_row_velocity(self) -> int:
        now_ns = monotonic_ns()
        last_ns = self._last_row0_ns
        last_row0 = self._last_row0_sample
        delta = self.row0 - last_row0 if last_ns is not None else 0

        if last_ns is not None and now_ns > last_ns:
            if delta:
                self._row0_velocity_samples.append((abs(delta), now_ns - last_ns))
                self._last_velocity_event_ns = now_ns
            elif (
                self._last_velocity_event_ns is not None
                and now_ns - self._last_velocity_event_ns > 750_000_000
            ):
                self._row0_velocity_samples.clear()
                self._last_velocity_event_ns = None

        self._last_row0_sample = self.row0
        self._last_row0_ns = now_ns
        return now_ns

    def _estimate_overscan_from_velocity(self, now_ns: int) -> int | None:
        if not self._row0_velocity_samples:
            return None

        last_event = self._last_velocity_event_ns
        if last_event is not None and now_ns - last_event > 750_000_000:
            self._row0_velocity_samples.clear()
            self._last_velocity_event_ns = None
            return None

        total_dt = sum(dt for _, dt in self._row0_velocity_samples)
        if total_dt <= 0:
            return None

        total_delta = sum(delta for delta, _ in self._row0_velocity_samples)
        rows_per_ns = total_delta / total_dt
        if rows_per_ns <= 0:
            return None

        rows_per_second = rows_per_ns * 1_000_000_000
        lookahead_seconds = 0.35
        hint = int(rows_per_second * lookahead_seconds)
        if hint <= 0:
            return None

        max_hint = max(0, self.view_height) * 10 or hint
        return max(0, min(hint, max_hint))

    def get_visible_table_slice(
        self, columns: Sequence[str], overscan_hint: int | None = None
    ) -> TableSlice:
        """Return the current viewport slice as an engine-agnostic table."""

        now_ns = self._update_row_velocity()
        if overscan_hint is None:
            overscan_hint = self._estimate_overscan_from_velocity(now_ns)

        budget_hint = self._frame_budget_overscan_hint
        self._frame_budget_overscan_hint = None
        if budget_hint is not None:
            if overscan_hint is None:
                overscan_hint = budget_hint
            else:
                overscan_hint = min(overscan_hint, budget_hint)

        return self._row_cache.get_visible_table_slice(columns, overscan_hint)

    def request_frame_budget_overscan(self, hint: int | None) -> None:
        """Limit overscan for the next visible slice fetch."""

        if hint is None:
            self._frame_budget_overscan_hint = None
            return
        try:
            value = int(hint)
        except (TypeError, ValueError):  # pragma: no cover - defensive
            self._frame_budget_overscan_hint = None
            return
        self._frame_budget_overscan_hint = max(0, value)

    def get_visible_dataframe(self, columns: Sequence[str]) -> pl.DataFrame:
        """Return the current viewport slice as a Polars DataFrame."""

        table_slice = self.get_visible_table_slice(columns)
        if not table_slice.columns:
            return pl.DataFrame()

        data = {
            column.name: pl.Series(column.name, list(column.values), dtype=column.dtype)
            for column in table_slice.columns
        }
        return pl.DataFrame(data)

    def update_terminal_metrics(self) -> None:
        # Start from actual terminal size via the injected UI hooks.
        prev_width = getattr(self, "view_width_chars", 0)
        prev_height = getattr(self, "view_height", 0)
        prev_prefetch = self._row_cache.prefetch

        try:
            cols, rows = self._ui_hooks.get_terminal_size((100, 30))
        except Exception:
            cols, rows = NullViewerUIHooks().get_terminal_size((100, 30))

        # Width estimate per column override only affects rendering width hints
        if self._viewport_cols_override is not None:
            cols = max(20, self._viewport_cols_override * 12)

        if self._view_width_override_chars is not None:
            cols = max(20, self._view_width_override_chars)
        elif is_test_mode():
            cols = max(20, 100)

        # Reserve lines for table header, separator, and status bar.
        # Required: header (1) + header separator (1) + empty line before status (1)
        # + status bar (1) + margin (1)
        reserved = 5
        base_view_height = max(1, rows - reserved)
        if self._viewport_rows_override is not None:
            # Never exceed the available height; honor override within bounds.
            view_height = max(1, min(base_view_height, self._viewport_rows_override))
        else:
            view_height = base_view_height
        view_width = max(20, cols)

        self.view_height = view_height
        prev_width_chars = getattr(self, "view_width_chars", None)
        self.view_width_chars = view_width
        self._autosized_widths.clear()
        # Re-apply maximize mode widths against the new terminal size
        self._apply_width_mode()

        new_prefetch = max(self.view_height * 4, 64)
        self._row_cache.prefetch = new_prefetch

        prefetch_changed = prev_prefetch is None or new_prefetch != prev_prefetch
        if view_width != prev_width or view_height != prev_height or prefetch_changed:
            self.invalidate_row_cache()
        if prev_width_chars is None or prev_width_chars != view_width:
            self.mark_status_dirty()

    def configure_terminal(self, width: int, height: int | None = None) -> None:
        """Configure explicit terminal metrics for deterministic renders."""

        clamped_width = max(20, int(width))
        if height is not None:
            self._viewport_rows_override = max(1, int(height))
        self.set_view_width_override(clamped_width)
        self.update_terminal_metrics()
        self.view_width_chars = clamped_width
        if height is not None:
            self.view_height = max(1, int(height))
        self._visible_key = None
        self.acknowledge_status_rendered()

    def set_view_width_override(self, width: int | None) -> None:
        """Force a specific character width for test or headless rendering."""

        if width is None:
            self._view_width_override_chars = None
        else:
            self._view_width_override_chars = max(20, int(width))

    def mark_status_dirty(self) -> None:
        """Signal that the status bar should be re-rendered."""

        self._status_dirty = True

    def acknowledge_status_rendered(self) -> None:
        """Mark the status bar as in sync with the latest render."""

        self._status_dirty = False

    def is_status_dirty(self) -> bool:
        """Return whether the status bar needs to be re-rendered."""

        return self._status_dirty

    @property
    def status_message(self) -> str | None:
        """Return the current status message displayed in the footer."""

        return self._status_message

    @status_message.setter
    def status_message(self, message: str | None) -> None:
        """Set the status message and mark the footer dirty if it changed."""

        if message == self._status_message:
            return
        self._status_message = message
        self.mark_status_dirty()

    def clamp(self) -> None:
        """Clamp the cursor and viewport to valid ranges."""

        before = (self.cur_row, self.cur_col, self.row0, self.col0)
        self._state.clamp()
        after = (self.cur_row, self.cur_col, self.row0, self.col0)
        if after != before:
            self.mark_status_dirty()

    def invalidate_row_count(self) -> None:
        """Mark the cached row count as stale."""
        self._row_counts.invalidate()

    def _ensure_total_rows(self) -> int | None:
        """Ensure we have an up-to-date total row count and return it."""
        return self._row_counts.ensure_total_rows()

    def goto_col(self, name: str) -> bool:
        """Move cursor to the column with the given name."""
        try:
            idx = self.columns.index(name)
        except ValueError:
            self.status_message = f"unknown column '{name}'"
            return False

        self.cur_col = idx
        if self.columns[idx] not in self.visible_cols and not self._is_column_frozen(idx):
            self.col0 = idx
        self.clamp()
        return True

    def _create_transformation_manager(self, sheet: Sheet) -> ViewerTransformationManager:
        """Build a transformation manager bound to ``sheet``."""

        sheet_for_history: SupportsSnapshots | None
        if hasattr(sheet, "snapshot_transforms") and hasattr(sheet, "restore_transforms"):
            sheet_for_history = sheet  # type: ignore[assignment]
        else:
            sheet_for_history = None
        return ViewerTransformationManager(
            history=TransformationHistory(sheet_for_history),
            capture_view_state=self._capture_view_state,
            restore_view_state=self._restore_view_state,
        )

    def _apply_plan_update(
        self, description: str, builder: Callable[[QueryPlan], QueryPlan]
    ) -> ChangeResult | None:
        """Apply a pure plan update produced by ``builder``."""

        return self._plan_controller.apply_plan_update(description, builder)

    def _status_from_error(self, operation: str, error: PulkaCoreError) -> None:
        """Format ``error`` into a viewer status message."""

        if isinstance(error, PlanError):
            category = "plan"
        elif isinstance(error, CompileError):
            category = "compile"
        elif isinstance(error, MaterializeError):
            category = "materialize"
        elif isinstance(error, CancelledError):
            category = "cancelled"
        else:
            category = "internal"

        detail = str(error).strip()
        if detail:
            message = f"{operation} {category} error: {detail}"
        else:
            message = f"{operation} {category} error"
        self.status_message = message[:120]

    def toggle_sort(self, col_name: str | None = None) -> None:
        """Toggle sort on the specified column (defaults to current column)."""
        if not self.columns:
            self.status_message = "no columns to sort"
            return

        if not sheet_supports(self.sheet, SHEET_FEATURE_PLAN):
            self.status_message = "sorting not supported"
            return

        target = col_name or self.columns[self.cur_col]
        try:
            result = self._apply_plan_update(
                f"sort {target}", lambda plan: plan_toggle_sort(plan, target)
            )
        except PulkaCoreError as exc:
            self._status_from_error("sort", exc)
            return
        except Exception as exc:
            self.status_message = f"sort error: {exc}"[:120]
            return

        if result is None:
            self.status_message = "sorting not supported"
            return

        if not result.plan_changed:
            self.status_message = "sort unchanged"
            return

        # Reset navigation after sort to mirror TUI behaviour
        self.cur_row = 0
        self.row0 = 0
        self.invalidate_row_count()

        plan = self._current_plan()
        sort_entries: tuple[tuple[str, bool], ...] = ()
        if plan is not None:
            sort_entries = tuple(plan.sort)

        if not sort_entries:
            self.status_message = "sort cleared"
        else:
            names = [name for name, _ in sort_entries]
            if target not in names:
                self.status_message = f"sort {target} cleared"
            else:
                primary, desc = sort_entries[0]
                suffix = "" if len(sort_entries) == 1 else f" ({len(sort_entries)} keys)"
                self.status_message = f"sort {primary} {'↓' if desc else '↑'}{suffix}"

        self.clamp()

    def apply_filter(self, filter_text: str | None) -> None:
        """Apply a filter expression to the active sheet."""

        if not sheet_supports(self.sheet, SHEET_FEATURE_PLAN):
            self.status_message = "filtering not supported"
            return

        normalized = None
        if filter_text is not None:
            stripped = filter_text.strip()
            normalized = stripped or None

        if normalized is not None:
            try:
                self.engine.validate_filter_clause(normalized)
            except PlanError as exc:
                detail = str(exc).strip()
                message = f"filter error: {detail}" if detail else "filter error"
                self.status_message = message[:120]
                return
            except Exception as exc:
                self.status_message = f"filter error: {exc}"[:120]
                return

        try:
            result = self._apply_plan_update(
                "filter change", lambda plan: plan_set_filter(plan, filter_text)
            )
        except PulkaCoreError as exc:
            self._status_from_error("filter", exc)
            return
        except Exception as exc:
            self.status_message = f"filter error: {exc}"[:120]
            return

        if result is None:
            self.status_message = "filtering not supported"
            return

        if not result.plan_changed:
            self.status_message = "filter unchanged"
            return
        self.cur_row = 0
        self.row0 = 0
        self.invalidate_row_count()
        active = self.filter_text
        if active:
            preview = active if len(active) <= 60 else active[:57] + "..."
            self.status_message = f"filter: {preview}"
        else:
            self.status_message = "filter cleared"
        self.clamp()

    def apply_sql_filter(self, where_clause: str | None) -> None:
        """Apply an SQL WHERE-clause filter to the active sheet."""

        if not sheet_supports(self.sheet, SHEET_FEATURE_PLAN):
            self.status_message = "SQL filtering not supported"
            return

        normalized_clause: str | None = None
        if where_clause is not None:
            trimmed = where_clause.strip()
            normalized_clause = trimmed or None

        if normalized_clause is not None:
            try:
                self.engine.validate_sql_where(self.sheet, normalized_clause)
            except PlanError as exc:
                detail = str(exc).strip()
                message = f"sql filter error: {detail}" if detail else "sql filter error"
                self.status_message = message[:120]
                return
            except Exception as exc:
                self.status_message = f"sql filter error: {exc}"[:120]
                return

        try:
            result = self._apply_plan_update(
                "sql filter change",
                lambda plan: plan_set_sql_filter(plan, normalized_clause),
            )
        except PulkaCoreError as exc:
            self._status_from_error("sql filter", exc)
            return
        except Exception as exc:
            self.status_message = f"sql filter error: {exc}"[:120]
            return

        if result is None:
            self.status_message = "SQL filtering not supported"
            return

        if not result.plan_changed:
            self.status_message = "filter unchanged"
            return
        self.cur_row = 0
        self.row0 = 0
        self.invalidate_row_count()
        active = self.filter_text
        if active:
            preview = active if len(active) <= 60 else active[:57] + "..."
            self.status_message = preview
        else:
            self.status_message = "filter cleared"
        self.clamp()

    def reset_filters(self) -> None:
        """Reset filters and sorting on the active sheet."""

        if not sheet_supports(self.sheet, SHEET_FEATURE_PLAN):
            self.status_message = "reset not supported"
            return

        try:
            result = self._apply_plan_update("reset filters", lambda _: plan_reset())
        except PulkaCoreError as exc:
            self._status_from_error("reset", exc)
            return
        except Exception as exc:
            self.status_message = f"reset error: {exc}"[:120]
            return

        if result is None:
            self.status_message = "reset not supported"
            return

        if not result.plan_changed:
            self.status_message = "filters already reset"
            return
        self.cur_row = 0
        self.row0 = 0
        self.invalidate_row_count()
        self.status_message = "filters reset"
        self._local_filter_text = None
        self._local_filter_kind = None
        self.search_text = None
        self.clamp()

    def set_search(self, text: str | None) -> None:
        """Record the active search text (whitespace-trimmed)."""
        cleaned = None if text is None else text.strip() or None
        current = self.search_text
        if cleaned == current:
            self.status_message = "search unchanged"
            return

        self.search_text = cleaned
        term = self.search_text
        if term:
            self.status_message = f"search: '{term}'"
        else:
            self.status_message = "search cleared"

    def _search_values(self, column: str) -> list[str]:
        """Return stringified column values for search operations."""
        total_rows = self._ensure_total_rows()
        if total_rows is None or total_rows <= 0:
            return []

        try:
            table_slice = self.sheet.fetch_slice(0, total_rows, [column])
        except Exception:
            return []

        if column not in table_slice.column_names:
            return []

        column_slice = table_slice.column(column)
        values: list[str] = []
        for val in column_slice.values:
            if val is None:
                values.append("")
            elif isinstance(val, str):
                values.append(val)
            else:
                try:
                    values.append(str(val))
                except Exception:
                    values.append("")
        return values

    def search(
        self,
        *,
        forward: bool,
        include_current: bool = False,
        center: bool = True,
        wrap: bool = True,
    ) -> bool:
        """Search within the current column for the recorded search string."""
        term = self.search_text
        self.search_text = term
        if not term:
            self.status_message = "search: no active query"
            return False

        if not self.columns:
            self.status_message = f"search '{term}': no columns"
            return False

        values = self._search_values(self.columns[self.cur_col])
        total_rows = len(values)
        if total_rows == 0:
            self.status_message = f"search '{term}': no rows"
            return False

        term_lower = term.lower()
        start_idx = max(0, min(self.cur_row, total_rows - 1))

        def _iter_positions() -> list[tuple[int, bool]]:
            if wrap:
                if forward:
                    first_segment = (
                        total_rows - start_idx
                        if include_current
                        else max(0, total_rows - (start_idx + 1))
                    )
                    idx = start_idx if include_current else start_idx + 1
                    step = 1
                else:
                    first_segment = start_idx + 1 if include_current else max(0, start_idx)
                    idx = start_idx if include_current else start_idx - 1
                    step = -1

                positions: list[tuple[int, bool]] = []
                current = idx
                for i in range(total_rows):
                    current %= total_rows
                    positions.append((current, i >= first_segment))
                    current += step
                return positions

            if forward:
                start = start_idx if include_current else start_idx + 1
                if start < 0:
                    start = 0
                if start >= total_rows:
                    return []
                return [(row, False) for row in range(start, total_rows)]
            start = start_idx if include_current else start_idx - 1
            if start < 0:
                return []
            if start >= total_rows:
                start = total_rows - 1
            return [(row, False) for row in range(start, -1, -1)]

        for row, wrapped in _iter_positions():
            value = values[row]
            if term_lower in value.lower():
                self.cur_row = row
                body_height = self._body_view_height()
                frozen_min = self._effective_frozen_row_count()
                if center:
                    half = max(1, body_height) // 2
                    self.row0 = max(frozen_min, self.cur_row - half)
                else:
                    if self.cur_row < self.row0:
                        self.row0 = max(frozen_min, self.cur_row)
                    elif self.cur_row >= self.row0 + body_height:
                        self.row0 = max(frozen_min, self.cur_row - body_height + 1)
                self.clamp()
                wrap_msg = " (wrapped)" if wrapped else ""
                self.status_message = f"search '{term}'{wrap_msg}"
                return True

        if wrap:
            self.status_message = f"search '{term}': no match"
        else:
            direction = "next" if forward else "previous"
            self.status_message = f"search '{term}': no {direction} match"
        return False

    def next_search_match(self) -> bool:
        """Advance to the next row search match."""

        return self.search(forward=True, include_current=False, center=True, wrap=False)

    def prev_search_match(self) -> bool:
        """Move to the previous row search match."""

        return self.search(forward=False, include_current=False, center=True, wrap=False)

    def replace_sheet(self, sheet: Sheet, *, source_path: str | None = None) -> None:
        """Swap the viewer to operate on a new sheet instance."""
        old_sheet = getattr(self, "sheet", None)
        if old_sheet is not None and old_sheet is not sheet:
            old_id = getattr(old_sheet, "sheet_id", None)
            preserve_id = getattr(sheet, "_preserve_jobs_from", None)
            if old_id is not None and old_id != preserve_id and self._runner is not None:
                with contextlib.suppress(Exception):
                    self._runner.invalidate_sheet(old_id)

        self.sheet = sheet
        self.columns = list(sheet.columns)
        self._schema_cache = getattr(sheet, "schema", {})
        self._source_path = source_path
        self._transformations = self._create_transformation_manager(sheet)
        provider = getattr(sheet, "row_provider", None)
        if provider is None:
            provider = RowProvider.for_sheet(sheet, runner=self._runner)
        elif getattr(provider, "_runner", None) is None:
            with contextlib.suppress(Exception):
                provider._runner = self._runner  # type: ignore[attr-defined]
        self._row_provider = provider
        self._engine = ViewerEngine(self._row_provider)
        self.cur_row = 0
        self.row0 = 0
        self.cur_col = 0
        self.col0 = 0
        self.invalidate_row_count()
        self._hidden_cols.clear()
        self._invalidate_frozen_columns_cache()
        self._header_widths = self._compute_initial_column_widths()
        self._default_header_widths = list(self._header_widths)
        self._width_mode = "default"
        self._width_target = None
        self._invalidate_width_cache()
        self._visible_key = None
        self._visible_cols_cached = self.columns[:1] if self.columns else []
        self.is_freq_view = False
        self.freq_source_col = None
        self.is_hist_view = False
        self.status_message = None
        self.update_terminal_metrics()
        self.clamp()
        self._reconcile_schema_changes()
        self._apply_sheet_freeze_defaults(sheet)
        self._apply_file_browser_layout_defaults(sheet)

    def replace_data(self, new_lf: pl.LazyFrame, *, source_path: str | None = None) -> None:
        """Compatibility helper that swaps in a new LazyFrame via DataSheet."""
        new_sheet = DataSheet(new_lf, runner=self._runner)
        self.replace_sheet(new_sheet, source_path=source_path)

    def _compute_initial_column_widths(self) -> list[int]:
        """Compute initial column widths based on header and sample data."""
        width_getter = getattr(self.sheet, "get_column_widths", None)
        if callable(width_getter):
            try:
                provided = width_getter()
            except Exception:
                provided = None
            if isinstance(provided, Mapping):
                widths: list[int] = []
                for name in self.columns:
                    raw = provided.get(name)
                    if isinstance(raw, int):
                        widths.append(max(self._min_col_width, raw))
                    else:
                        widths.append(max(self._min_col_width, len(name) + 2))
                if widths:
                    return widths
        return self._widths.compute_initial_widths()

    def _compute_content_width(self, col_idx: int) -> int:
        """Compute the maximum content width for a column by sampling data."""
        if col_idx < 0 or col_idx >= len(self.columns):
            return self._min_col_width

        col_name = self.columns[col_idx]

        # Start with header width
        max_width = len(col_name) + 2

        try:
            # Sample some data to estimate content width
            sample_rows = min(1000, self._total_rows if self._total_rows else 100)

            # Get current visible columns to minimize fetch cost
            visible_cols = self.visible_cols
            sample_cols = (col_name,) if col_name not in visible_cols else tuple(visible_cols)

            # Fetch sample data
            sample_slice = self.sheet.fetch_slice(0, sample_rows, sample_cols)

            if col_name in sample_slice.column_names and sample_slice.height > 0:
                column = sample_slice.column(col_name)

                try:
                    formatted_values = column.formatted(0)
                except Exception:
                    formatted_values = None

                if formatted_values:
                    for raw_value, rendered in zip(column.values, formatted_values, strict=False):
                        if raw_value is None or rendered == "":
                            display = "null"
                        elif isinstance(raw_value, float) and (
                            math.isnan(raw_value) or math.isinf(raw_value)
                        ):
                            if math.isnan(raw_value):
                                display = "NaN"
                            else:
                                display = "inf" if raw_value > 0 else "-inf"
                        else:
                            display = rendered
                        max_width = max(max_width, len(display) + 2)
                else:
                    lengths = [len(str(value)) for value in column.values if value is not None]
                    if lengths:
                        max_width = max(max_width, max(lengths) + 2)

        except Exception:
            # If anything goes wrong with sampling, fall back to header width
            pass

        # Cap at reasonable limits
        result = max(self._min_col_width, min(max_width, 200))  # Cap at 200 chars max
        return result

    def _invalidate_width_cache(self) -> None:
        """Drop cached content width calculations."""
        self._widths.invalidate_cache()

    def _ensure_default_widths(self) -> None:
        """Ensure default header widths align with the current schema."""
        self._widths.ensure_default_widths()

    def _normalize_width_mode(self) -> None:
        """Validate current width mode against the active columns."""
        self._widths.normalize_mode()

    def _apply_width_mode(self) -> None:
        """Rebuild header widths according to the active width mode."""
        self._widths.apply_width_mode()

    def _autosize_visible_columns(self, column_indices: list[int]) -> None:
        """Stretch visible column widths to fill the viewport when possible."""
        self._widths.autosize_visible_columns(column_indices)

    def _force_default_width_mode(self) -> None:
        """Reset width mode to default and reapply widths."""
        self._widths.force_default_mode()

    def toggle_maximize_current_col(self) -> None:
        """Toggle maximisation for the active column."""
        self._widths.toggle_maximize_current_col()

    def toggle_maximize_all_cols(self) -> None:
        """Toggle maximisation for every column."""
        self._widths.toggle_maximize_all_cols()

    # Width mode helpers exposed for other components ---------------------------------

    @property
    def maximized_column_index(self) -> int | None:
        """Return the index of the maximised column when in single-column mode."""
        if self._width_mode == "single" and self._width_target is not None:
            return self._width_target
        return None

    @property
    def all_columns_maximized(self) -> bool:
        """Return True when all columns are currently maximised."""
        return self._width_mode == "all"

    @property
    def width_mode_state(self) -> dict[str, int | None | str]:
        """Return a serialisable representation of the active width mode."""
        return {"mode": self._width_mode, "target": self.maximized_column_index}

    def snapshot(self) -> ViewerPublicState:
        """Return an immutable snapshot of the viewer suitable for public use."""

        columns = tuple(self.columns)
        visible_columns = tuple(self.visible_cols)
        hidden_columns = tuple(name for name in columns if name in self._hidden_cols)
        frozen_columns = tuple(self.frozen_columns)

        view_height = getattr(self, "view_height", 0)
        visible_row_count = max(0, view_height)
        visible_column_count = len(visible_columns)

        col_span = visible_column_count
        row_span = visible_row_count

        viewport = ViewerViewport(
            row0=self.row0,
            rowN=self.row0 + max(0, row_span - 1),
            col0=self.col0,
            colN=self.col0 + max(0, col_span - 1),
            rows=row_span,
            cols=col_span,
        )

        highlighted_column: str | None = None
        if columns and 0 <= self.cur_col < len(columns):
            highlighted_column = columns[self.cur_col]

        return ViewerPublicState(
            cursor=ViewerCursor(row=self.cur_row, col=self.cur_col),
            viewport=viewport,
            columns=columns,
            visible_columns=visible_columns,
            hidden_columns=hidden_columns,
            frozen_columns=frozen_columns,
            total_rows=self._total_rows,
            visible_row_count=visible_row_count,
            total_columns=len(columns),
            visible_column_count=visible_column_count,
            hidden_column_count=len(hidden_columns),
            status_message=self.status_message,
            highlighted_column=highlighted_column,
            width_mode=self._width_mode,
            width_target=self.maximized_column_index,
            all_columns_maximized=self.all_columns_maximized,
            sort_column=self.sort_col,
            sort_ascending=self.sort_asc,
            filter_text=self.filter_text,
            filter_kind=self.filter_kind,
            search_text=self.search_text,
            frequency_mode=self.is_freq_view,
            frequency_source_column=self.freq_source_col,
        )

    # Navigation helpers (shared with keybindings and scripted mode)
    def move_down(self, steps: int = 1):
        for _ in range(max(steps, 0)):
            # Stop at the last row if we know the total
            if (
                self._total_rows is not None
                and self._total_rows > 0
                and self.cur_row >= self._total_rows - 1
            ):
                break
            self.cur_row += 1
            if self.cur_row >= self.row0 + self._body_view_height():
                # Don't scroll past the end of the data
                if self._total_rows is not None and self._total_rows > 0:
                    max_row0 = self._max_row0_for_total(self._total_rows)
                    self.row0 = min(self.row0 + 1, max_row0)
                else:
                    self.row0 += 1
        self.clamp()

    def move_up(self, steps: int = 1):
        for _ in range(max(steps, 0)):
            self.cur_row = max(0, self.cur_row - 1)
            self.row0 = min(self.row0, self.cur_row)
        self.clamp()

    def move_right(self, steps: int = 1):
        """
        Move cursor right following predictable behavior:
        1. Move cursor to next visible column
        2. If cursor is not visible after move, shift viewport just enough to make it visible
        3. Always move cursor to next column (no jumps), ensuring sequential progression
        """
        # Clear any previous max visible column constraint for normal navigation
        self._max_visible_col = None
        self._visible_key = None

        for _ in range(max(steps, 0)):
            # Find next visible column
            next_col = self.cur_col + 1
            while next_col < len(self.columns) and self.columns[next_col] in self._hidden_cols:
                next_col += 1

            # Stop if we've reached the last column
            if next_col >= len(self.columns):
                break

            # Check if target column is currently visible
            target_col_visible = self.columns[next_col] in self.visible_cols

            # Move cursor to next column
            self.cur_col = next_col

            # CRITICAL: Ensure cursor is never behind the viewport
            # This prevents the bug where cursor becomes invisible
            if self.cur_col < self.col0:
                self.col0 = self.cur_col

            # If we're moving to a column that wasn't visible, position it as rightmost
            if not target_col_visible:
                self._ensure_col_visible_right_aligned(self.cur_col)

        self.clamp()

    def move_left(self, steps: int = 1):
        # Clear any previous max visible column constraint for normal navigation
        self._max_visible_col = None
        self._visible_key = None

        for _ in range(max(steps, 0)):
            # Find previous visible column
            prev_col = self.cur_col - 1
            while prev_col >= 0 and self.columns[prev_col] in self._hidden_cols:
                prev_col -= 1
            if prev_col >= 0:
                self.cur_col = prev_col
                # Adjust viewport if needed
                visible_cols = self.visible_cols
                if (
                    visible_cols
                    and self.columns[self.cur_col] not in visible_cols
                    and not self._is_column_frozen(self.cur_col)
                ):
                    # Adjust col0 to make current column visible
                    self.col0 = min(self.col0, self.cur_col)
        self.clamp()

    def page_down(self):
        step = self._body_view_height()
        if self._total_rows is not None and self._total_rows > 0:
            self.cur_row = min(self.cur_row + step, self._total_rows - 1)
            self.row0 = min(self.row0 + step, self._max_row0_for_total(self._total_rows))
        else:
            self.cur_row += step
            self.row0 += step
        self.clamp()

    def page_up(self):
        step = self._body_view_height()
        self.cur_row = max(0, self.cur_row - step)
        self.row0 = max(0, self.row0 - step)
        self.clamp()

    def go_top(self):
        self.cur_row = 0
        self.row0 = 0
        self.clamp()

    def go_bottom(self):
        # Jump to the known last row if we have a count; otherwise best-effort
        total_rows = self._ensure_total_rows()
        if total_rows is not None and total_rows > 0:
            self.cur_row = total_rows - 1
            self.row0 = self._max_row0_for_total(total_rows)
        else:
            self.cur_row += 1_000_000
            self.row0 = max(self.row0, self.cur_row - self._body_view_height() + 1)
        self.clamp()

    def first_col(self):
        # Go to the leftmost visible column
        visible = self.visible_cols
        if visible:
            try:
                self.cur_col = self.columns.index(visible[0])
            except ValueError:
                self.cur_col = max(0, self.col0)
        else:
            self.cur_col = max(0, self.col0)
        self.clamp()

    def last_col(self):
        # Go to rightmost visible column
        visible = self.visible_cols
        if visible:
            try:
                self.cur_col = self.columns.index(visible[-1])
            except ValueError:
                self.cur_col = max(0, self.col0)
        else:
            self.cur_col = max(0, self.col0)
        self.clamp()

    def first_col_overall(self):
        # Go to very first column overall
        self.cur_col = 0
        # Ensure the first column is visible by adjusting viewport
        self.col0 = min(self.col0, self.cur_col)
        self.clamp()

    def _ensure_col_visible_right_aligned(self, target: int) -> None:
        """Make sure `target` is visible as the RIGHTMOST visible column."""
        should_record = self._perf_callback is not None
        start_ns = monotonic_ns() if should_record else 0
        iterations = 0
        start_col0 = self.col0

        self._visible_key = None

        # Set the max visible column constraint to ensure target is rightmost
        self._max_visible_col = target

        if self._is_column_frozen(target):
            # Frozen columns are always visible; no need to adjust scroll state
            if should_record:
                duration_ms = (monotonic_ns() - start_ns) / 1_000_000
                payload = {
                    "target": target,
                    "iterations": 0,
                    "start_col0": start_col0,
                    "final_col0": self.col0,
                    "viewport_override": bool(self._viewport_cols_override),
                    "frozen": True,
                }
                self._record_perf_event("viewer.ensure_right_aligned", duration_ms, payload)
            return

        # If user forced a fixed number of visible columns, this is trivial.
        if self._viewport_cols_override:
            max_cols = max(1, self._viewport_cols_override)
            new_col0 = max(0, target - max_cols + 1)
            self.col0 = max(self._first_scrollable_col_index(), new_col0)
            if should_record:
                duration_ms = (monotonic_ns() - start_ns) / 1_000_000
                payload = {
                    "target": target,
                    "iterations": 0,
                    "start_col0": start_col0,
                    "final_col0": self.col0,
                    "viewport_override": True,
                }
                self._record_perf_event("viewer.ensure_right_aligned", duration_ms, payload)
            return

        max_width = max(20, self.view_width_chars)
        frozen_indices = self._frozen_column_indices()
        frozen_set = self._frozen_column_index_set()

        used_width = 1  # table border on the left
        for idx in frozen_indices:
            name = self.columns[idx]
            if name in self._hidden_cols:
                continue
            used_width += self._header_widths[idx] + 1

        new_col0 = target
        idx = target

        # Always account for the target column even if it exceeds the viewport on its own.
        while idx >= 0:
            iterations += 1
            name = self.columns[idx]
            if name in self._hidden_cols or idx in frozen_set:
                idx -= 1
                continue

            width_with_border = self._header_widths[idx] + 1
            if idx == target:
                used_width += width_with_border
                new_col0 = idx
                idx -= 1
                continue

            if used_width + width_with_border > max_width:
                break

            used_width += width_with_border
            new_col0 = idx
            idx -= 1

        computed_col0 = max(self._first_scrollable_col_index(), new_col0)
        if target >= start_col0:
            computed_col0 = max(computed_col0, start_col0)
        self.col0 = computed_col0

        if should_record:
            duration_ms = (monotonic_ns() - start_ns) / 1_000_000
            payload = {
                "target": target,
                "iterations": iterations,
                "start_col0": start_col0,
                "final_col0": self.col0,
                "viewport_override": False,
            }
            self._record_perf_event("viewer.ensure_right_aligned", duration_ms, payload)

    def last_col_overall(self):
        self.cur_col = len(self.columns) - 1
        self._ensure_col_visible_right_aligned(self.cur_col)
        self.clamp()

    def slide_column_left(self) -> None:
        """Slide the current column one visible slot to the left."""
        self._slide_current_column("left", to_edge=False)

    def slide_column_right(self) -> None:
        """Slide the current column one visible slot to the right."""
        self._slide_current_column("right", to_edge=False)

    def slide_column_to_start(self) -> None:
        """Slide the current column to the first visible position."""
        self._slide_current_column("left", to_edge=True)

    def slide_column_to_end(self) -> None:
        """Slide the current column to the last visible position."""
        self._slide_current_column("right", to_edge=True)

    def _slide_current_column(self, direction: Literal["left", "right"], *, to_edge: bool) -> None:
        if not self.columns:
            self.status_message = "no columns to move"
            return

        current_idx = self.cur_col
        if not (0 <= current_idx < len(self.columns)):
            self.status_message = "column index out of range"
            return

        current_name = self.columns[current_idx]
        if current_name in self._hidden_cols:
            self.status_message = "cannot move hidden column"
            return

        visible_indices = [
            idx for idx, name in enumerate(self.columns) if name not in self._hidden_cols
        ]
        if len(visible_indices) <= 1:
            self.status_message = "no other visible columns"
            return

        try:
            visible_pos = visible_indices.index(current_idx)
        except ValueError:
            self.status_message = "column is not visible"
            return

        if direction == "left":
            if visible_pos == 0:
                self.status_message = "already at left edge"
                return
            target_visible_pos = 0 if to_edge else visible_pos - 1
            target_idx = visible_indices[target_visible_pos]
            insert_at = target_idx
        else:
            last_pos = len(visible_indices) - 1
            if visible_pos == last_pos:
                self.status_message = "already at right edge"
                return
            target_visible_pos = last_pos if to_edge else visible_pos + 1
            target_idx = visible_indices[target_visible_pos]
            insert_at = target_idx + 1

        new_order = list(self.columns)
        removed = new_order.pop(current_idx)
        if target_idx > current_idx:
            insert_at -= 1
        insert_at = max(0, min(insert_at, len(new_order)))
        new_order.insert(insert_at, removed)

        if new_order == list(self.columns):
            self.status_message = "column already positioned"
            return

        if to_edge and direction == "left":
            desc = f"move {current_name} to first column"
            status = f"{current_name} moved to first column"
        elif to_edge and direction == "right":
            desc = f"move {current_name} to last column"
            status = f"{current_name} moved to last column"
        elif direction == "left":
            desc = f"move {current_name} left"
            status = f"{current_name} moved left"
        else:
            desc = f"move {current_name} right"
            status = f"{current_name} moved right"

        def mutate() -> bool:
            self._apply_column_reorder(new_order, active_column=current_name)
            return True

        result = self._transformations.record_change(desc, mutate)
        if result.committed:
            self.status_message = status

    def _apply_column_reorder(self, new_order: Sequence[str], *, active_column: str) -> None:
        old_order = tuple(self.columns)
        if tuple(new_order) == old_order:
            return

        index_lookup = {name: idx for idx, name in enumerate(old_order)}

        def _reorder_widths(values: Sequence[int]) -> list[int]:
            if not values:
                return [self._min_col_width] * len(new_order)
            reordered: list[int] = []
            for name in new_order:
                idx = index_lookup.get(name)
                if idx is None or idx >= len(values):
                    reordered.append(self._min_col_width)
                else:
                    reordered.append(values[idx])
            return reordered

        self.columns = list(new_order)
        self._header_widths = _reorder_widths(self._header_widths)
        self._default_header_widths = _reorder_widths(self._default_header_widths)

        width_target_name: str | None = None
        if (
            self._width_mode == "single"
            and self._width_target is not None
            and 0 <= self._width_target < len(old_order)
        ):
            width_target_name = old_order[self._width_target]

        self._invalidate_width_cache()
        self._autosized_widths.clear()
        self._visible_key = None
        self._visible_cols_cached = self.columns[:1] if self.columns else []
        self._max_visible_col = None
        self._invalidate_frozen_columns_cache()
        self.invalidate_row_cache()

        if width_target_name is not None and width_target_name in self.columns:
            self._width_target = self.columns.index(width_target_name)
        elif self._width_mode == "single":
            self._width_mode = "default"
            self._width_target = None

        if active_column in self.columns:
            self.cur_col = self.columns.index(active_column)
        else:
            self.cur_col = 0

        self.clamp()

        if self.columns:
            active_name = self.columns[self.cur_col]
            if active_name not in self.visible_cols and not self._is_column_frozen(self.cur_col):
                self._ensure_col_visible_right_aligned(self.cur_col)
                self.clamp()

    def center_current_row(self):
        """Center the current row in the viewport."""
        half = max(1, self._body_view_height()) // 2
        self.row0 = max(self._effective_frozen_row_count(), self.cur_row - half)
        self.clamp()

    def _get_current_value(self):
        """Get the value at the current cursor position."""
        try:
            # Fetch a single row at current position
            current_col = self.columns[self.cur_col]
            table_slice = self.sheet.fetch_slice(self.cur_row, 1, [current_col])
            if (
                table_slice.height > 0
                and current_col in table_slice.column_names
                and table_slice.column(current_col).values
            ):
                return table_slice.column(current_col).values[0]
            return None
        except Exception:
            return None

    def prev_different_value(self):
        """Navigate to the previous row with a different value in the current column."""
        if self.cur_row <= 0:
            self.status_message = "already at top"
            return False

        current_value = self._get_current_value()
        current_col = self.columns[self.cur_col]

        try:
            # Use cached slices to scan for the previous different value.
            start_idx = max(0, self.cur_row - 1000)  # Look at most 1000 rows back
            table_slice = self.sheet.fetch_slice(start_idx, self.cur_row - start_idx, [current_col])

            if current_col not in table_slice.column_names:
                raise ValueError("column not in slice")

            values = table_slice.column(current_col).values

            # Search backward in the fetched data for different value
            for i in range(table_slice.height - 1, -1, -1):  # from last to first
                if values[i] != current_value:
                    # Found a different value
                    target_row = start_idx + i
                    target_value = values[i]
                    self.cur_row = target_row
                    self.clamp()
                    self.status_message = f"found different value: {target_value}"
                    return True

        except Exception:
            pass

        self.status_message = "no different value found above"
        return False

    def next_different_value(self):
        """Navigate to the next row with a different value in the current column."""
        current_value = self._get_current_value()
        current_col = self.columns[self.cur_col]

        try:
            # Use cached slices to scan forward for the next different value.
            table_slice = self.sheet.fetch_slice(
                self.cur_row + 1, 1000, [current_col]
            )  # Look at most 1000 rows ahead

            if current_col not in table_slice.column_names:
                raise ValueError("column not in slice")

            values = table_slice.column(current_col).values

            # Search forward in the fetched data for different value
            for i, value in enumerate(values):
                if value != current_value:
                    # Found a different value
                    target_row = self.cur_row + 1 + i
                    target_value = value
                    self.cur_row = target_row
                    self.clamp()
                    self.status_message = f"found different value: {target_value}"
                    return True

        except Exception:
            pass

        self.status_message = "no different value found below"
        return False

    @property
    def visible_cols(self) -> list[str]:
        # Determine visible columns based on new dynamic width allocation logic
        visible_columns = [col for col in self.columns if col not in self._hidden_cols]
        frozen_indices = self._frozen_column_indices()
        frozen_names = self.frozen_columns
        frozen_set = self._frozen_column_name_set()

        # Handle viewport override early (fixed number of columns regardless of width).
        if self._viewport_cols_override is not None:
            should_record = self._perf_callback is not None
            start_ns = monotonic_ns() if should_record else 0

            max_dynamic = max(1, self._viewport_cols_override)
            scroll_start = max(self.col0, self._first_scrollable_col_index())
            dynamic: list[str] = []
            for idx in range(scroll_start, len(self.columns)):
                name = self.columns[idx]
                if name in self._hidden_cols or name in frozen_set:
                    continue
                dynamic.append(name)
                if len(dynamic) >= max_dynamic:
                    break

            result = frozen_names + dynamic
            self._last_col_fits_completely = True
            self._visible_cols_cached = result
            self._visible_key = None

            index_lookup = {name: idx for idx, name in enumerate(self.columns)}
            visible_indices = [index_lookup[col] for col in result if col in index_lookup]
            self._autosize_visible_columns(visible_indices)

            if should_record:
                duration_ms = (monotonic_ns() - start_ns) / 1_000_000
                payload = {
                    "viewport_override": True,
                    "col0": self.col0,
                    "visible_count": len(result),
                    "hidden_count": len(self._hidden_cols),
                    "frozen": len(frozen_names),
                }
                self._record_perf_event("viewer.visible_cols", duration_ms, payload)
            return result

        key = (
            self.col0,
            self.cur_col,
            self.view_width_chars,
            self._viewport_cols_override,
            tuple(sorted(self._hidden_cols)),
            tuple(self._header_widths),
            tuple(frozen_indices),
        )
        if key == self._visible_key:
            return self._visible_cols_cached

        should_record = self._perf_callback is not None
        start_ns = monotonic_ns() if should_record else 0
        evaluated = 0

        max_width = max(20, self.view_width_chars)
        allow_partial_width = self._width_mode == "default"
        used = 1  # left border
        res: list[str] = []
        last_fits_completely = True

        for idx in frozen_indices:
            name = self.columns[idx]
            if name in self._hidden_cols:
                continue
            res.append(name)
            used += self._header_widths[idx] + 1

        if used > max_width:
            last_fits_completely = False

        scroll_start = max(self.col0, self._first_scrollable_col_index())

        for idx in range(scroll_start, len(self.columns)):
            name = self.columns[idx]
            if name in self._hidden_cols or name in frozen_set:
                continue

            evaluated += 1
            w = self._header_widths[idx]

            if used + w + 1 <= max_width:
                used += w + 1
                res.append(name)
            else:
                if len(res) == len(frozen_names):
                    res.append(name)
                    last_fits_completely = False
                elif allow_partial_width:
                    min_needed = self._min_col_width + 1
                    if used + min_needed <= max_width:
                        res.append(name)
                        last_fits_completely = False
                break

        active_name: str | None = None
        if 0 <= self.cur_col < len(self.columns):
            active_name = self.columns[self.cur_col]

        active_is_partial_last = (
            allow_partial_width
            and not last_fits_completely
            and not self._aligning_active_column
            and self._viewport_cols_override is None
            and res
            and active_name is not None
            and res[-1] == active_name
        )

        if active_is_partial_last:
            prev_col0 = self.col0
            self._aligning_active_column = True
            try:
                self._ensure_col_visible_right_aligned(self.cur_col)
            finally:
                self._aligning_active_column = False
            if self.col0 != prev_col0:
                self._visible_key = None
                return self.visible_cols

        if not res and visible_columns:
            for col in visible_columns:
                col_idx = self.columns.index(col)
                if col_idx >= scroll_start:
                    res = [col]
                    break
            if not res:
                res = [visible_columns[0]]

        if self._max_visible_col is not None and res:
            truncated_res = []
            for col in res:
                col_idx = self.columns.index(col)
                if col in frozen_set or col_idx <= self._max_visible_col:
                    truncated_res.append(col)
                else:
                    break
            res = truncated_res

        index_lookup = {name: idx for idx, name in enumerate(self.columns)}
        visible_indices = [index_lookup[col] for col in res if col in index_lookup]
        self._autosize_visible_columns(visible_indices)

        self._visible_key = key
        self._visible_cols_cached = res
        self._last_col_fits_completely = last_fits_completely

        if should_record:
            duration_ms = (monotonic_ns() - start_ns) / 1_000_000
            payload = {
                "viewport_override": False,
                "col0": self.col0,
                "visible_count": len(res),
                "evaluated": evaluated,
                "hidden_count": len(self._hidden_cols),
                "max_width": max_width,
                "fits_full": last_fits_completely,
                "max_visible_col": self._max_visible_col,
                "frozen": len(frozen_names),
            }
            self._record_perf_event("viewer.visible_cols", duration_ms, payload)
        return res

    # Column hiding functionality
    def hide_current_column(self) -> None:
        """Hide the current column (- key)."""
        if not self.columns:
            self.status_message = "No columns to hide"
            return

        current_col_name = self.columns[self.cur_col]

        # If column is already hidden, no-op
        if current_col_name in self._hidden_cols:
            self.status_message = f"Column already hidden: {current_col_name}"
            return

        # If this is the last visible column, block the operation
        visible_columns = self.visible_columns()
        if len(visible_columns) == 1 and visible_columns[0] == current_col_name:
            self.status_message = f"Cannot hide the last visible column: {current_col_name}"
            return

        def builder(plan: QueryPlan) -> QueryPlan:
            base_projection = list(plan.projection_or(self.columns))
            if current_col_name not in base_projection:
                return plan
            updated = [name for name in base_projection if name != current_col_name]
            if not updated:
                return plan
            return plan_set_projection(plan, updated)

        try:
            result = self._apply_plan_update(f"hide {current_col_name}", builder)
        except PulkaCoreError as exc:
            self._status_from_error("hide column", exc)
            return

        if result is None:

            def mutate() -> bool:
                self._local_hidden_cols.add(current_col_name)
                self._update_hidden_column_cache(set(self._local_hidden_cols))
                return True

            result = self._transformations.record_change(f"hide {current_col_name}", mutate)
            if not result.committed:
                return
        elif not result.committed:
            return

        self.status_message = f"Removed column: {current_col_name}"
        self._visible_key = None
        self._reconcile_schema_changes()

        if __debug__:
            self._validate_state_consistency()

    def unhide_all_columns(self) -> None:
        """Unhide all hidden columns (gv command)."""
        # If no hidden columns, no-op
        if not self._hidden_cols:
            self.status_message = "No hidden columns to restore"
            return

        def builder(plan: QueryPlan) -> QueryPlan:
            return plan_set_projection(plan, self.columns)

        try:
            result = self._apply_plan_update("unhide all columns", builder)
        except PulkaCoreError as exc:
            self._status_from_error("unhide", exc)
            return

        if result is None:

            def mutate() -> bool:
                self._local_hidden_cols.clear()
                self._update_hidden_column_cache(set())
                return True

            result = self._transformations.record_change("unhide all columns", mutate)
            if not result.committed:
                return
        elif not result.committed:
            return

        self.status_message = "Restored all hidden columns"
        self._visible_key = None
        self._reconcile_schema_changes()

        if __debug__:
            self._validate_state_consistency()

    def undo_last_operation(self) -> None:
        """Undo last operation (u key)."""

        snapshot = self._transformations.undo()
        if snapshot is None:
            self.status_message = "Nothing to undo"
            return
        if snapshot.description:
            self.status_message = f"Undo: {snapshot.description}"
        else:
            self.status_message = "Undid last change"

        self.search_text = self.search_text

        if __debug__:
            self._validate_state_consistency()

    def redo_last_operation(self) -> None:
        """Redo last undone operation (U key)."""

        snapshot = self._transformations.redo()
        if snapshot is None:
            self.status_message = "Nothing to redo"
            return
        if snapshot.description:
            self.status_message = f"Redo: {snapshot.description}"
        else:
            self.status_message = "Redid last change"

        self.search_text = self.search_text

        if __debug__:
            self._validate_state_consistency()

    def visible_columns(self) -> list[str]:
        """Return the ordered list of columns marked visible by the active plan."""

        projection = self._plan_projection_columns()
        if projection is not None:
            order_index = {name: idx for idx, name in enumerate(self.columns)}
            filtered = [name for name in projection if name in order_index]
            filtered.sort(key=lambda name: order_index.get(name, len(self.columns)))
            return filtered
        return [col for col in self.columns if col not in self._hidden_cols]

    def current_colname(self) -> str:
        """Return the name of the current column."""
        return self.columns[self.cur_col]

    def _capture_view_state(self) -> ViewerSnapshot:
        """Capture viewer-specific state required for undo/redo."""

        return self._state.capture_snapshot()

    def _restore_view_state(self, state: ViewerSnapshot) -> None:
        """Restore viewer state from ``state`` and reconcile caches."""

        self._state.restore_snapshot(state)

    def next_visible_col_index(self, search_from: int) -> int | None:
        """Find the next visible column index, searching rightward, else leftward as fallback."""
        return self._state.next_visible_col_index(search_from)

    def _move_cursor_to_next_visible_column(self) -> None:
        """Move cursor to the next visible column."""
        self._state.move_cursor_to_next_visible_column()

    def _ensure_cursor_on_visible_column(self) -> None:
        """Ensure cursor is positioned on a visible column."""
        self._state.ensure_cursor_on_visible_column()

    def _update_hidden_column_cache(self, hidden: set[str], *, ensure_cursor: bool = True) -> None:
        """Apply ``hidden`` as the canonical hidden column set."""

        normalized = {name for name in hidden if name in self.columns}
        self._hidden_cols = normalized
        self._invalidate_frozen_columns_cache()
        self._visible_key = None

        if len(self._header_widths) < len(self.columns):
            for idx in range(len(self._header_widths), len(self.columns)):
                baseline = (
                    self._default_header_widths[idx]
                    if idx < len(self._default_header_widths)
                    else self._min_col_width
                )
                self._header_widths.append(max(baseline, self._min_col_width))

        for idx, name in enumerate(self.columns):
            if idx >= len(self._header_widths):
                break
            if name in normalized:
                self._header_widths[idx] = 0
            elif self._header_widths[idx] == 0:
                baseline = (
                    self._default_header_widths[idx]
                    if idx < len(self._default_header_widths)
                    else self._min_col_width
                )
                self._header_widths[idx] = max(baseline, self._min_col_width)

        if ensure_cursor:
            self._ensure_cursor_on_visible_column()

    def _sync_hidden_columns_from_plan(self) -> None:
        """Align hidden columns with the authoritative projection source."""

        if not self.columns:
            self._local_hidden_cols.clear()
            self._update_hidden_column_cache(set())
            return

        projection = self._plan_projection_columns()
        if projection is None:
            self._local_hidden_cols.intersection_update(self.columns)
            self._update_hidden_column_cache(set(self._local_hidden_cols))
            return

        projected = set(projection)
        hidden = {name for name in self.columns if name not in projected}
        self._local_hidden_cols.clear()
        self._update_hidden_column_cache(hidden)

    def _reconcile_schema_changes(self) -> None:
        """
        Reconcile internal state after schema changes:
        - Reconcile hidden set with current schema
        - Rebuild column widths to align with current schema
        - Invalidate all caches related to visibility and rendering
        - Re-snap cursor to the nearest visible column
        - Clean undo stack by dropping references to removed columns
        """
        # Refresh cached schema snapshot
        self._schema_cache = getattr(self.sheet, "schema", self._schema_cache)

        # Align hidden caches with the authoritative plan/local projection
        self._sync_hidden_columns_from_plan()
        self._local_hidden_cols.intersection_update(self.columns)

        # Rebuild column widths to align with current schema
        # Ensure widths array matches current columns
        if len(self._header_widths) != len(self.columns):
            # Rebuild widths arrays to match current schema
            self._header_widths = self._compute_initial_column_widths()
            self._default_header_widths = list(self._header_widths)
            self._force_default_width_mode()
        else:
            self._apply_width_mode()

        # Invalidate all visibility-related caches
        self._visible_key = None
        self._visible_cols_cached = self.columns[:1] if self.columns else []

        # Re-snap cursor to the nearest visible column
        # Cursor must always land on a visible column
        if self.columns:
            self.cur_col = max(0, min(self.cur_col, len(self.columns) - 1))
            # Ensure cursor is on a visible column
            if self.cur_col < len(self.columns) and self.columns[self.cur_col] in self._hidden_cols:
                # Try to find next visible column
                next_visible = self.next_visible_col_index(self.cur_col)
                if next_visible is not None:
                    self.cur_col = next_visible
                else:
                    # If no next visible, find first visible column
                    for i, col in enumerate(self.columns):
                        if col not in self._hidden_cols:
                            self.cur_col = i
                            break
                    else:
                        # Emergency fallback - reset to first column
                        self.cur_col = 0

        # Prune transformation history snapshots referencing stale columns
        column_set = set(self.columns)

        def _is_valid(snapshot: TransformationSnapshot) -> bool:
            state = snapshot.view_state
            if not isinstance(state, ViewerSnapshot):
                return True
            if len(state.header_widths) != len(self.columns):
                return False
            return set(state.hidden_cols).issubset(column_set)

        self._transformations.filter_history(_is_valid)

        # Ensure at least one column remains visible
        visible_columns = self.visible_columns()
        if not visible_columns and self.columns and self._plan_projection_columns() is None:
            # Only legacy, planless sheets can end up hiding everything.
            self._local_hidden_cols.clear()
            self._update_hidden_column_cache(set())
            self._force_default_width_mode()

    def _validate_state_consistency(self) -> None:
        """Validate that internal state is consistent.

        This method checks for common consistency issues that could lead to
        UI synchronization problems.
        """
        # Ensure cursor is on a valid column
        if not (0 <= self.cur_col < len(self.columns)):
            raise AssertionError(
                f"Cursor column index {self.cur_col} out of bounds [0, {len(self.columns)})"
            )

        # Ensure cursor is on a visible column (unless all hidden, shouldn't happen)
        if self.columns and self.columns[self.cur_col] in self._hidden_cols:
            raise AssertionError(f"Cursor is on hidden column {self.columns[self.cur_col]}")

        missing_hidden = self._hidden_cols - set(self.columns)
        if missing_hidden:
            raise AssertionError(f"Hidden columns out of schema: {sorted(missing_hidden)}")

        missing_local = self._local_hidden_cols - set(self.columns)
        if missing_local:
            raise AssertionError(f"Local hidden columns out of schema: {sorted(missing_local)}")

        projection = self._plan_projection_columns()
        if projection is not None:
            expected_hidden = {name for name in self.columns if name not in set(projection)}
            if expected_hidden != self._hidden_cols:
                raise AssertionError(
                    "Plan projection mismatch:"
                    f" expected {sorted(expected_hidden)}, got {sorted(self._hidden_cols)}"
                )
            expected_visible = [name for name in projection if name in self.columns]
            if expected_visible != self.visible_columns():
                raise AssertionError(
                    "Plan projection mismatch:",
                    f" visible {expected_visible} != {self.visible_columns()}",
                )

        # Ensure header widths array matches columns length
        if len(self._header_widths) != len(self.columns):
            raise AssertionError(
                f"Header widths mismatch: got {len(self._header_widths)}, want {len(self.columns)}"
            )



from __future__ import annotations

import polars as pl

from pulka.core.viewer import Viewer
from pulka.core.viewer.public_state import viewer_public_state
from pulka.core.viewer.row_count_tracker import RowCountTracker
from pulka.core.viewer.state import ViewerSnapshot
from pulka.core.viewer.ui_hooks import NullViewerUIHooks
from pulka.sheets.data_sheet import DataSheet


def _make_plan_viewer(job_runner) -> Viewer:
    df = pl.DataFrame({"id": [1, 2, 3], "name": ["a", "b", "c"]})
    sheet = DataSheet(df.lazy(), runner=job_runner)
    return Viewer(sheet, viewport_rows=5, viewport_cols=5, runner=job_runner)


class _LegacySheet:
    def __init__(self) -> None:
        self.columns = ["id", "name"]
        self.schema = {"id": pl.Int64, "name": pl.Utf8}

    def fetch_slice(self, row_start: int, row_count: int, columns: list[str]) -> pl.DataFrame:
        return pl.DataFrame({name: [] for name in columns})


def _make_legacy_viewer(job_runner) -> Viewer:
    sheet = _LegacySheet()
    return Viewer(sheet, viewport_rows=5, viewport_cols=5, runner=job_runner)


def test_state_snapshot_roundtrip_restores_hidden_columns(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    state = viewer.state_controller

    baseline_widths = tuple(viewer._header_widths)
    viewer._local_hidden_cols = {"name"}
    viewer._update_hidden_column_cache(set(viewer._local_hidden_cols))
    viewer.cur_row = 2
    viewer.row0 = 1
    viewer.cur_col = 0
    viewer.col0 = 0

    snapshot = state.capture_snapshot()

    viewer._hidden_cols.clear()
    viewer._local_hidden_cols.clear()
    viewer._header_widths = [1, 1]
    viewer.cur_row = 0
    viewer.row0 = 0

    state.restore_snapshot(snapshot)

    assert viewer._hidden_cols == {"name"}
    assert viewer._local_hidden_cols == {"name"}
    assert tuple(viewer._header_widths) == baseline_widths
    assert viewer.cur_row == 2
    assert viewer.row0 == 1


def test_state_clamp_skips_hidden_columns(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    viewer.cur_col = viewer.columns.index("name")
    viewer._local_hidden_cols = {"name"}
    viewer._update_hidden_column_cache(set(viewer._local_hidden_cols), ensure_cursor=False)

    viewer.clamp()

    assert viewer.columns[viewer.cur_col] == "id"


def test_state_restore_trims_unknown_hidden_columns_and_extends_widths(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    controller = viewer.state_controller

    viewer.columns.append("extra")
    viewer._default_header_widths.append(12)
    viewer._header_widths = [8, 9]  # shorter than columns on purpose
    viewer.cur_row = 0
    viewer.row0 = 0

    snapshot = ViewerSnapshot(
        hidden_cols=("missing",),
        header_widths=(5,),
        cur_col=0,
        col0=0,
        cur_row=2,
        row0=1,
    )

    controller.restore_snapshot(snapshot)

    assert viewer._hidden_cols == set()
    assert viewer._local_hidden_cols == set()
    assert len(viewer._header_widths) == len(viewer.columns)
    assert viewer._header_widths[-1] >= viewer._min_col_width


def test_viewer_public_snapshot_exposes_sanitised_state(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    viewer.cur_row = 1
    viewer.row0 = 1
    viewer.cur_col = viewer.columns.index("name")
    viewer.col0 = 0
    viewer._local_hidden_cols = {"name"}
    viewer._update_hidden_column_cache(set(viewer._local_hidden_cols), ensure_cursor=False)

    state = viewer.snapshot()

    assert isinstance(state.visible_columns, tuple)
    assert isinstance(state.hidden_columns, tuple)
    assert state.cursor.row == 1
    assert state.cursor.col == viewer.columns.index("name")
    assert state.viewport.row0 == 1
    assert state.hidden_column_count == len(state.hidden_columns)
    assert "name" in state.hidden_columns
    assert "name" not in state.visible_columns


def test_viewer_public_state_helper_handles_missing_snapshot(job_runner):
    viewer = _make_legacy_viewer(job_runner)

    state = viewer_public_state(viewer)

    assert state is not None
    assert state.cursor.row == viewer.cur_row

    class _LegacyViewer:
        def __init__(self) -> None:
            self.cur_row = 0
            self.cur_col = 1
            self.row0 = 2
            self.col0 = 3
            self.columns = ["x", "y"]
            self.visible_cols = ["x"]
            self._hidden_cols = {"y"}
            self.sort_col = "x"
            self.sort_asc = True

    legacy = _LegacyViewer()

    legacy_state = viewer_public_state(legacy)

    assert legacy_state is not None
    assert legacy_state.cursor.row == 0
    assert legacy_state.cursor.col == 1
    assert legacy_state.viewport.row0 == 2
    assert legacy_state.visible_columns == ("x",)
    assert legacy_state.hidden_columns == ("y",)


def test_reconcile_schema_changes_respects_plan_projection(job_runner):
    viewer = _make_plan_viewer(job_runner)
    viewer.hide_current_column()
    viewer._reconcile_schema_changes()

    projection = tuple(viewer.sheet.plan.projection_or(viewer.columns))
    assert projection == tuple(viewer.visible_columns())
    assert viewer._local_hidden_cols == set()


def test_reconcile_schema_changes_unhides_everything_for_planless_sheet(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    viewer._local_hidden_cols = set(viewer.columns)
    viewer._update_hidden_column_cache(set(viewer._local_hidden_cols))

    viewer._reconcile_schema_changes()

    assert set(viewer.visible_columns()) == set(viewer.columns)
    assert viewer._local_hidden_cols == set()


def test_plan_controller_apply_plan_update_sets_limit(job_runner):
    viewer = _make_plan_viewer(job_runner)
    controller = viewer.plan_controller

    result = controller.apply_plan_update("limit", lambda plan: plan.with_limit(1))

    assert result is not None
    assert result.plan_changed is True
    assert viewer.sheet.plan.limit == 1


class _DummySheet:
    sheet_id = None

    def __len__(self) -> int:  # pragma: no cover - trivial
        return 8


class _DummyViewer:
    def __init__(self, runner) -> None:
        self.sheet = _DummySheet()
        self._total_rows: int | None = None
        self._row_count_stale = True
        self._row_count_future = None
        self._row_count_display_pending = False
        self._status_dirty = False
        self._ui_hooks = NullViewerUIHooks()
        self.invalidate_called = False
        self.job_runner = runner

    def invalidate_row_cache(self) -> None:
        self.invalidate_called = True

    def clamp(self) -> None:  # pragma: no cover - exercised indirectly
        pass

    @property
    def ui_hooks(self) -> NullViewerUIHooks:
        return self._ui_hooks

    def mark_status_dirty(self) -> None:
        self._status_dirty = True

    def acknowledge_status_rendered(self) -> None:
        self._status_dirty = False

    def is_status_dirty(self) -> bool:
        return self._status_dirty


def test_row_count_tracker_invalidates_and_counts(job_runner):
    dummy = _DummyViewer(job_runner)
    tracker = RowCountTracker(dummy, runner=job_runner)

    tracker.invalidate()

    assert dummy.is_status_dirty() is True
    assert dummy.invalidate_called is True
    assert dummy._total_rows is None

    total = tracker.ensure_total_rows()

    assert total == 8
    assert tracker.total_rows == 8
    assert dummy._total_rows == 8
    assert dummy._row_count_stale is False
    assert dummy._row_count_display_pending is False
    assert dummy.is_status_dirty() is True


def test_configure_terminal_clears_status_and_resets_cache(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    viewer.mark_status_dirty()
    viewer._visible_key = (1, 2, 3)
    viewer.configure_terminal(72, 10)

    assert viewer.view_width_chars == 72
    assert viewer.view_height == 10
    assert viewer._visible_key is None
    assert viewer.is_status_dirty() is False


def test_acknowledge_status_rendered_resets_flag(job_runner):
    viewer = _make_legacy_viewer(job_runner)
    viewer.mark_status_dirty()
    assert viewer.is_status_dirty() is True

    viewer.acknowledge_status_rendered()

    assert viewer.is_status_dirty() is False
